try:
    from .opengl import GLRasterizer
except ImportError:
    pass
from .cpu import (rasterize_barycentric_coordinates,
                  rasterize_barycentric_coordinate_images)
from .base import (rasterize_mesh_from_barycentric_coordinate_images,
                   rasterize_shape_image_from_barycentric_coordinate_images,
                   rasterize_mesh)
from .transform import model_to_clip_transform, clip_to_image_transform



import warnings
from functools import partial

from menpo.base import partial_doc
from menpo.io.input.base import (glob_with_suffix, _import_glob_lazy_list,
                                 _import, _import_object_attach_landmarks,
                                 _data_dir_path, _data_path_to,
                                 _ls_builtin_assets, BuiltinAssets,
                                 _import_builtin_asset,
                                 _register_importer)
from menpo.io.input import same_name, image_paths
from menpo3d.base import menpo3d_src_dir_path
from .extensions import mesh_types, mesh_landmark_types, lsfm_types


def mesh_paths(pattern):
    r"""
    Return mesh filepaths that Menpo3d can import that match the glob pattern.
    """
    return glob_with_suffix(pattern, mesh_types)


def landmark_file_paths(pattern):
    r"""
    Return landmark file filepaths that Menpo3d can import that match the glob
    pattern.
    """
    return glob_with_suffix(pattern, mesh_landmark_types)


def same_name_texture(path, paths_callable=image_paths):
    r"""
    Default image texture resolver. Returns **the lexicographically
    sorted first** texture found to have the same stem as the asset. A warning
    is raised if more than one texture is found.
    """
    # pattern finding all landmarks with the same stem
    pattern = path.with_suffix('.*')
    texture_paths = sorted(paths_callable(pattern))
    if len(texture_paths) > 1:
        warnings.warn('More than one texture found for file, returning '
                      'only the first.')
    if not texture_paths:
        return None
    return texture_paths[0]


same_name_landmark = partial_doc(same_name, paths_callable=landmark_file_paths)

menpo3d_data_dir_path = partial_doc(_data_dir_path, menpo3d_src_dir_path)

menpo3d_ls_builtin_assets = partial_doc(_ls_builtin_assets, 
                                        menpo3d_data_dir_path)

menpo3d_data_path_to = partial_doc(_data_path_to, menpo3d_data_dir_path,
                                   menpo3d_ls_builtin_assets)

_menpo3d_import_builtin_asset = partial_doc(_import_builtin_asset,
                                            menpo3d_data_path_to,
                                            mesh_types, mesh_landmark_types,
                                            texture_resolver=same_name_texture)


register_mesh_importer = partial_doc(_register_importer, mesh_types)
register_landmark_importer = partial_doc(_register_importer, 
                                         mesh_landmark_types)


def import_mesh(filepath, landmark_resolver=same_name_landmark,
                texture_resolver=same_name_texture):
    r"""Single mesh (and associated landmarks and texture) importer.

    Iff an mesh file is found at `filepath`, returns a :map:`TriMesh`
    representing it. Landmark files sharing the same filename
    will be imported and attached too. If texture coordinates and a suitable
    texture are found the object returned will be a :map:`TexturedTriMesh`.

    Parameters
    ----------
    filepath : `str`
        A relative or absolute filepath to an image file.
    landmark_resolver : `function`, optional
        This function will be used to find landmarks for the
        mesh. The function should take one argument (the mesh itself) and
        return a dictionary of the form ``{'group_name': 'landmark_filepath'}``
        Default finds landmarks with the same name as the mesh file.
    texture : `bool`, optional
        If ``False``, don't search for textures.

    Returns
    -------
    trimesh : :map:`TriMesh`
        An instantiated :map:`TriMesh` (or subclass thereof)
    """
    kwargs = {'texture_resolver': texture_resolver}
    return _import(filepath, mesh_types,
                   landmark_resolver=landmark_resolver,
                   landmark_ext_map=mesh_landmark_types,
                   landmark_attach_func=_import_object_attach_landmarks,
                   importer_kwargs=kwargs)


def import_meshes(pattern, max_meshes=None, shuffle=False,
                  landmark_resolver=same_name, textures=True,
                  as_generator=False, verbose=False):
    r"""Multiple mesh importer.

    Makes it's best effort to import and attach relevant related
    information such as landmarks. It searches the directory for files that
    begin with the same filename and end in a supported extension.

    If texture coordinates and a suitable texture are found the object
    returned will be a :map:`TexturedTriMesh`.

    Note that this is a function returns a :map:`LazyList`. Therefore, the
    function will return immediately and indexing into the returned list
    will load the landmarks at run time. If all meshes should be loaded, then
    simply wrap the returned :map:`LazyList` in a Python `list`.

    Parameters
    ----------
    pattern : `str`
        The glob path pattern to search for textures and meshes.
    max_meshes : positive `int`, optional
        If not ``None``, only import the first ``max_meshes`` meshes found.
        Else, import all.
    shuffle : `bool`, optional
        If ``True``, the order of the returned meshes will be randomised. If
        ``False``, the order of the returned meshes will be alphanumerically
        ordered.
    landmark_resolver : `function`, optional
        This function will be used to find landmarks for the
        mesh. The function should take one argument (the mesh itself) and
        return a dictionary of the form ``{'group_name': 'landmark_filepath'}``
        Default finds landmarks with the same name as the mesh file.
    textures : `bool`, optional
        If ``False``, don't search for textures.
    as_generator : `bool`, optional
        If ``True``, the function returns a generator and assets will be yielded
        one after another when the generator is iterated over.
    verbose : `bool`, optional
        If ``True`` progress of the importing will be dynamically reported.

    Returns
    -------
    lazy_list : :map:`LazyList` or generator of Python objects
        A :map:`LazyList` or generator yielding Python objects inside the
        pickle files found to match the glob pattern provided.

    Raises
    ------
    ValueError
        If no meshes are found at the provided glob.
    """
    kwargs = {'texture': textures}
    return _import_glob_lazy_list(
        pattern, mesh_types, max_assets=max_meshes, shuffle=shuffle,
        landmark_resolver=landmark_resolver,
        landmark_ext_map=mesh_landmark_types,
        importer_kwargs=kwargs, as_generator=as_generator,
        landmark_attach_func=_import_object_attach_landmarks, verbose=verbose)


def import_landmark_file(filepath, landmark_resolver=same_name):
    r"""Single landmark group importer.

    Iff an landmark file is found at `filepath`, returns a :map:`LandmarkGroup`
    representing it.

    Parameters
    ----------
    filepath : `str`
        A relative or absolute filepath to an landmark file.

    Returns
    -------
    :map:`LandmarkGroup`
        The :map:`LandmarkGroup` that the file format represents.

    """
    return _import(filepath, mesh_landmark_types,
                   landmark_resolver=landmark_resolver)


def import_landmark_files(pattern, max_landmarks=None, shuffle=False,
                          as_generator=False, verbose=False):
    r"""Multiple landmark file importer.

    Note that this is a function returns a :map:`LazyList`. Therefore, the
    function will return immediately and indexing into the returned list
    will load the landmarks at run time. If all landmarks should be loaded, then
    simply wrap the returned :map:`LazyList` in a Python `list`.

    Parameters
    ----------
    pattern : `str`
        The glob path pattern to search for landmarks.
    max_landmarks : positive `int`, optional
        If not ``None``, only import the first ``max_landmarks`` found.
        Else, import all.
    shuffle : `bool`, optional
        If ``True``, the order of the returned landmark files will be
        randomised. If ``False``, the order of the returned landmark files will
        be alphanumerically ordered.
    as_generator : `bool`, optional
        If ``True``, the function returns a generator and assets will be yielded
        one after another when the generator is iterated over.
    verbose : `bool`, optional
        If ``True`` progress of the importing will be dynamically reported.

    Returns
    ------
    lazy_list : :map:`LazyList` or generator of Python objects
        A :map:`LazyList` or generator yielding Python objects inside the
        pickle files found to match the glob pattern provided.

    Raises
    ------
    ValueError
        If no landmarks are found at the provided glob.

    """
    return _import_glob_lazy_list(pattern, mesh_landmark_types,
                                  max_assets=max_landmarks,
                                  shuffle=shuffle, as_generator=as_generator,
                                  verbose=verbose)


def import_lsfm_model(filepath):
    r"""Import a LSFM Morphable Model


    Parameters
    ----------
    filepath : `str`
        A relative or absolute filepath to a LSFM model mat file.

    Returns
    -------
    :map:`PCAModel`
        The 3DMM contained in the LSFM mat file.

    """
    return _import(filepath, lsfm_types,
                   landmark_resolver=None)

import_builtin_asset = BuiltinAssets(_menpo3d_import_builtin_asset)

for asset in menpo3d_ls_builtin_assets():
    setattr(import_builtin_asset, asset.replace('.', '_'),
            partial(_menpo3d_import_builtin_asset, asset))



import numpy as np

from menpo.visualize import Renderer

# The colour map used for all lines and markers
GLOBAL_CMAP = 'jet'


def _parse_marker_size(marker_size, points):
    if marker_size is None:
        from menpo.shape import PointCloud
        pc = PointCloud(points, copy=False)
        # This is the way that mayavi automatically computes the scale factor in
        # case the user passes scale_factor = 'auto'. We use it for both the
        # marker_size as well as the numbers_size.
        xyz_min, xyz_max = pc.bounds()
        x_min, y_min, z_min = xyz_min
        x_max, y_max, z_max = xyz_max
        distance = np.sqrt(((x_max - x_min) ** 2 +
                            (y_max - y_min) ** 2 +
                            (z_max - z_min) ** 2) /
                           (4 * pc.n_points ** 0.33))
        if distance == 0:
            marker_size = 1
        else:
            marker_size = 0.1 * distance
    return marker_size


def _parse_colour(colour):
    from matplotlib.colors import ColorConverter
    return ColorConverter().to_rgb(colour)


def _check_colours_list(render_flag, colours_list, n_objects, error_str):
    from menpo.visualize.viewmatplotlib import sample_colours_from_colourmap
    if render_flag:
        if colours_list is None:
            # sample colours from jet colour map
            colours_list = sample_colours_from_colourmap(n_objects, GLOBAL_CMAP)
        if isinstance(colours_list, list):
            if len(colours_list) == 1:
                colours_list[0] = _parse_colour(colours_list[0])
                colours_list *= n_objects
            elif len(colours_list) != n_objects:
                raise ValueError(error_str)
        else:
            colours_list = [_parse_colour(colours_list)] * n_objects
    else:
        colours_list = [None] * n_objects
    return colours_list


def _set_numbering(figure, centers, render_numbering=True, numbers_size=None,
                   numbers_colour='k'):
    import mayavi.mlab as mlab
    numbers_colour = _parse_colour(numbers_colour)
    numbers_size = _parse_marker_size(numbers_size, centers)
    if render_numbering:
        for k, p in enumerate(centers):
            mlab.text3d(p[0], p[1], p[2], str(k), figure=figure,
                        scale=numbers_size, orient_to_camera=True,
                        color=numbers_colour, line_width=2)


class MayaviRenderer(Renderer):
    """
    Abstract class for performing visualizations using Mayavi.

    Parameters
    ----------
    figure_id : str or `None`
        A figure name or `None`. `None` assumes we maintain the Mayavi
        state machine and use `mlab.gcf()`.
    new_figure : bool
        If `True`, creates a new figure to render on.
    """
    def __init__(self, figure_id, new_figure):
        try:
            import mayavi.mlab as mlab
        except ImportError:
            raise ImportError("mayavi is required for viewing 3D objects "
                              "(consider 'conda/pip install mayavi')")
        super(MayaviRenderer, self).__init__(figure_id, new_figure)

        self._supported_ext = ['png', 'jpg', 'bmp', 'tiff',  # 2D
                               'ps', 'eps', 'pdf',  # 2D
                               'rib', 'oogl', 'iv', 'vrml', 'obj']  # 3D
        n_ext = len(self._supported_ext)
        func_list = [lambda obj, fp, **kwargs: mlab.savefig(fp.name, **obj)] * n_ext
        self._extensions_map = dict(zip(['.' + s for s in self._supported_ext],
                                    func_list))
        # To store actors for clearing
        self._actors = []

    def get_figure(self):
        r"""
        Gets the figure specified by the combination of `self.figure_id` and
        `self.new_figure`. If `self.figure_id == None` then `mlab.gcf()`
        is used. `self.figure_id` is also set to the correct id of the figure
        if a new figure is created.

        Returns
        -------
        figure : Mayavi figure object
            The figure we will be rendering on.
        """
        import mayavi.mlab as mlab
        if self.new_figure or self.figure_id is not None:
            self.figure = mlab.figure(self.figure_id, bgcolor=(1, 1, 1))
            # and reset the view to z forward, y up.
            self.figure.scene.camera.view_up = np.array([0, 1, 0])
        else:
            self.figure = mlab.gcf()

        self.figure_id = self.figure.name

        return self.figure

    def save_figure(self, filename, format='png', size=None,
                    magnification='auto', overwrite=False):
        r"""
        Method for saving the figure of the current `figure_id` to file.

        Parameters
        ----------
        filename : `str` or `file`-like object
            The string path or file-like object to save the figure at/into.
        format : `str`
            The format to use. This must match the file path if the file path is
            a `str`.
        size : `tuple` of `int` or ``None``, optional
            The size of the image created (unless magnification is set,
            in which case it is the size of the window used for rendering). If
            ``None``, then the figure size is used.
        magnification :	`double` or ``'auto'``, optional
            The magnification is the scaling between the pixels on the screen,
            and the pixels in the file saved. If you do not specify it, it will
            be calculated so that the file is saved with the specified size.
            If you specify a magnification, Mayavi will use the given size as a
            screen size, and the file size will be ``magnification * size``.
            If ``'auto'``, then the magnification will be set automatically.
        overwrite : `bool`, optional
            If ``True``, the file will be overwritten if it already exists.
        """
        from menpo.io.output.base import _export
        savefig_args = {'size': size, 'figure': self.figure,
                        'magnification': magnification}
        # Use the export code so that we have a consistent interface
        _export(savefig_args, filename, self._extensions_map, format,
                overwrite=overwrite)

    @property
    def width(self):
        r"""
        The width of the scene in pixels.

        :type: `int`
        """
        return self.figure.scene.get_size()[0]

    @property
    def height(self):
        r"""
        The height of the scene in pixels.

        :type: `int`
        """
        return self.figure.scene.get_size()[1]

    @property
    def modelview_matrix(self):
        r"""
        Retrieves the modelview matrix for this scene.

        :type: ``(4, 4)`` `ndarray`
        """
        camera = self.figure.scene.camera
        return camera.view_transform_matrix.to_array().astype(np.float32)

    @property
    def projection_matrix(self):
        r"""
        Retrieves the projection matrix for this scene.

        :type: ``(4, 4)`` `ndarray`
        """
        scene = self.figure.scene
        camera = scene.camera
        scene_size = tuple(scene.get_size())
        aspect_ratio = float(scene_size[0]) / float(scene_size[1])
        p = camera.get_projection_transform_matrix(
            aspect_ratio, -1, 1).to_array().astype(np.float32)
        return p

    @property
    def renderer_settings(self):
        r"""
        Returns all the information required to construct an identical
        renderer to this one.

        Returns
        -------
        settings : `dict`
            The dictionary with the following keys:

                * ``'width'`` (`int`) : The width of the scene.
                * ``'height'`` (`int`) : The height of the scene.
                * ``'model_matrix'`` (`ndarray`) : The model array (identity).
                * ``'view_matrix'`` (`ndarray`) : The view array.
                * ``'projection_matrix'`` (`ndarray`) : The projection array.

        """
        return {'width': self.width,
                'height': self.height,
                'model_matrix': np.eye(4, dtype=np.float32),
                'view_matrix': self.modelview_matrix,
                'projection_matrix': self.projection_matrix}

    def clear_figure(self):
        r"""
        Method for clearing the current figure.
        """
        from mayavi import mlab
        mlab.clf(figure=self.figure)
        if len(self._actors) > 0:
            self.figure.scene.remove_actors(self._actors)

    def force_draw(self):
        r"""
        Method for forcing the current figure to render. This is useful for
        the widgets animation.
        """
        from pyface.api import GUI
        _gui = GUI()
        orig_val = _gui.busy
        _gui.set_busy(busy=True)
        _gui.set_busy(busy=orig_val)
        _gui.process_events()


class MayaviVectorViewer3d(MayaviRenderer):
    def __init__(self, figure_id, new_figure, points, vectors):
        super(MayaviVectorViewer3d, self).__init__(figure_id, new_figure)
        self.points = points
        self.vectors = vectors

    def render(self, colour='r', line_width=2, marker_style='2darrow',
               marker_resolution=8, marker_size=None, step=None, alpha=1.0):
        from mayavi import mlab
        marker_size = _parse_marker_size(marker_size, self.points)
        colour = _parse_colour(colour)
        mlab.quiver3d(self.points[:, 0], self.points[:, 1], self.points[:, 2],
                      self.vectors[:, 0], self.vectors[:, 1], self.vectors[:, 2],
                      figure=self.figure, color=colour, mask_points=step,
                      line_width=line_width, mode=marker_style,
                      resolution=marker_resolution, opacity=alpha,
                      scale_factor=marker_size)
        return self


class MayaviPointGraphViewer3d(MayaviRenderer):
    def __init__(self, figure_id, new_figure, points, edges):
        super(MayaviPointGraphViewer3d, self).__init__(figure_id, new_figure)
        self.points = points
        self.edges = edges

    def render(self, render_lines=True, line_colour='r', line_width=2,
               render_markers=True, marker_style='sphere', marker_size=None,
               marker_colour='r', marker_resolution=8, step=None, alpha=1.0,
               render_numbering=False, numbers_colour='k', numbers_size=None):
        from mayavi import mlab

        # Render the lines if requested
        if render_lines:
            line_colour = _parse_colour(line_colour)
            # TODO: Make step work for lines as well
            # Create the points
            if step is None:
                step = 1
            src = mlab.pipeline.scalar_scatter(self.points[:, 0],
                                               self.points[:, 1],
                                               self.points[:, 2])
            # Connect them
            src.mlab_source.dataset.lines = self.edges
            # The stripper filter cleans up connected lines
            lines = mlab.pipeline.stripper(src)

            # Finally, display the set of lines
            mlab.pipeline.surface(lines, figure=self.figure, opacity=alpha,
                                  line_width=line_width, color=line_colour)

        # Render the markers if requested
        if render_markers:
            marker_size = _parse_marker_size(marker_size, self.points)
            marker_colour = _parse_colour(marker_colour)
            mlab.points3d(self.points[:, 0], self.points[:, 1],
                          self.points[:, 2], figure=self.figure,
                          scale_factor=marker_size, mode=marker_style,
                          color=marker_colour, opacity=alpha,
                          resolution=marker_resolution, mask_points=step)

        # set numbering
        _set_numbering(self.figure, self.points, numbers_size=numbers_size,
                       render_numbering=render_numbering,
                       numbers_colour=numbers_colour)

        return self


class MayaviTriMeshViewer3d(MayaviRenderer):
    def __init__(self, figure_id, new_figure, points, trilist):
        super(MayaviTriMeshViewer3d, self).__init__(figure_id, new_figure)
        self.points = points
        self.trilist = trilist

    def _render_mesh(self, mesh_type, line_width, colour, marker_size,
                     marker_resolution, marker_style, step, alpha):
        import mayavi.mlab as mlab
        marker_size = _parse_marker_size(marker_size, self.points)
        colour = _parse_colour(colour)
        mlab.triangular_mesh(self.points[:, 0], self.points[:, 1],
                             self.points[:, 2], self.trilist,
                             figure=self.figure, line_width=line_width,
                             representation=mesh_type, color=colour,
                             scale_factor=marker_size, mask_points=step,
                             resolution=marker_resolution, mode=marker_style,
                             opacity=alpha, tube_radius=None)

    def render(self, mesh_type='wireframe', line_width=2, colour='r',
               marker_style='sphere', marker_size=None, marker_resolution=8,
               normals=None, normals_colour='k', normals_line_width=2,
               normals_marker_style='2darrow', normals_marker_size=None,
               normals_marker_resolution=8, step=None, alpha=1.0):
        if normals is not None:
            MayaviVectorViewer3d(self.figure_id, False,
                                 self.points, normals).render(
                colour=normals_colour, line_width=normals_line_width, step=step,
                marker_style=normals_marker_style,
                marker_resolution=normals_marker_resolution,
                marker_size=normals_marker_size, alpha=alpha)
        self._render_mesh(mesh_type, line_width, colour, marker_size,
                          marker_resolution, marker_style, step, alpha)
        return self


class MayaviTexturedTriMeshViewer3d(MayaviRenderer):
    def __init__(self, figure_id, new_figure, points, trilist, texture,
                 tcoords_per_point):
        super(MayaviTexturedTriMeshViewer3d, self).__init__(figure_id,
                                                            new_figure)
        self.points = points
        self.trilist = trilist
        self.texture = texture
        self.tcoords_per_point = tcoords_per_point
        self._actors = []

    def _render_mesh(self, mesh_type='surface', ambient_light=0.0,
                     specular_light=0.0, alpha=1.0):
        from tvtk.api import tvtk
        pd = tvtk.PolyData()
        pd.points = self.points
        pd.polys = self.trilist
        pd.point_data.t_coords = self.tcoords_per_point
        mapper = tvtk.PolyDataMapper()
        mapper.set_input_data(pd)
        p = tvtk.Property(representation=mesh_type, opacity=alpha,
                          ambient=ambient_light, specular=specular_light)
        actor = tvtk.Actor(mapper=mapper, property=p)
        # Get the pixels from our image class which are [0, 1] and scale
        # back to valid pixels. Then convert to tvtk ImageData.
        texture = self.texture.pixels_with_channels_at_back(out_dtype=np.uint8)
        if self.texture.n_channels == 1:
            texture = np.stack([texture] * 3, axis=-1)
        image_data = np.flipud(texture).ravel()
        image_data = image_data.reshape([-1, 3])
        image = tvtk.ImageData()
        image.point_data.scalars = image_data
        image.dimensions = self.texture.width, self.texture.height, 1
        texture = tvtk.Texture()
        texture.set_input_data(image)
        actor.texture = texture
        self.figure.scene.add_actors(actor)
        self._actors.append(actor)

    def render(self, mesh_type='surface', ambient_light=0.0, specular_light=0.0,
               normals=None, normals_colour='k', normals_line_width=2,
               normals_marker_style='2darrow', normals_marker_resolution=8,
               normals_marker_size=None, step=None, alpha=1.0):
        if normals is not None:
            MayaviVectorViewer3d(self.figure_id, False,
                                 self.points, normals).render(
                colour=normals_colour, line_width=normals_line_width, step=step,
                marker_style=normals_marker_style,
                marker_resolution=normals_marker_resolution,
                marker_size=normals_marker_size, alpha=alpha)
        self._render_mesh(mesh_type=mesh_type, ambient_light=ambient_light,
                          specular_light=specular_light, alpha=alpha)
        return self


class MayaviColouredTriMeshViewer3d(MayaviRenderer):
    def __init__(self, figure_id, new_figure, points, trilist,
                 colour_per_point):
        super(MayaviColouredTriMeshViewer3d, self).__init__(figure_id,
                                                            new_figure)
        self.points = points
        self.trilist = trilist
        self.colour_per_point = colour_per_point
        self._actors = []

    def _render_mesh(self, mesh_type='surface', ambient_light=0.0,
                     specular_light=0.0, alpha=1.0):
        from tvtk.api import tvtk
        pd = tvtk.PolyData()
        pd.points = self.points
        pd.polys = self.trilist
        pd.point_data.scalars = (self.colour_per_point * 255.).astype(np.uint8)
        mapper = tvtk.PolyDataMapper()
        mapper.set_input_data(pd)
        p = tvtk.Property(representation=mesh_type, opacity=alpha,
                          ambient=ambient_light, specular=specular_light)
        actor = tvtk.Actor(mapper=mapper, property=p)
        self.figure.scene.add_actors(actor)
        self._actors.append(actor)

    def render(self, mesh_type='surface', ambient_light=0.0, specular_light=0.0,
               normals=None, normals_colour='k', normals_line_width=2,
               normals_marker_style='2darrow', normals_marker_resolution=8,
               normals_marker_size=None, step=None, alpha=1.0):
        if normals is not None:
            MayaviVectorViewer3d(self.figure_id, False,
                                 self.points, normals).render(
                colour=normals_colour, line_width=normals_line_width, step=step,
                marker_style=normals_marker_style,
                marker_resolution=normals_marker_resolution,
                marker_size=normals_marker_size, alpha=alpha)
        self._render_mesh(mesh_type=mesh_type, ambient_light=ambient_light,
                          specular_light=specular_light, alpha=alpha)
        return self


class MayaviSurfaceViewer3d(MayaviRenderer):
    def __init__(self, figure_id, new_figure, values, mask=None):
        super(MayaviSurfaceViewer3d, self).__init__(figure_id, new_figure)
        if mask is not None:
            values[~mask] = np.nan
        self.values = values

    def render(self, colour=(1, 0, 0), line_width=2, step=None,
               marker_style='2darrow', marker_resolution=8, marker_size=0.05,
               alpha=1.0):
        from mayavi import mlab
        warp_scale = kwargs.get('warp_scale', 'auto')
        mlab.surf(self.values, warp_scale=warp_scale)
        return self


class MayaviLandmarkViewer3d(MayaviRenderer):
    def __init__(self, figure_id, new_figure, group, landmark_group):
        super(MayaviLandmarkViewer3d, self).__init__(figure_id, new_figure)
        self.group = group
        self.landmark_group = landmark_group

    def render(self, render_lines=True, line_colour='r', line_width=2,
               render_markers=True, marker_style='sphere', marker_size=None,
               marker_colour='r', marker_resolution=8, step=None, alpha=1.0,
               render_numbering=False, numbers_colour='k', numbers_size=None):
        # Regarding the labels colours, we may get passed either no colours (in
        # which case we generate random colours) or a single colour to colour
        # all the labels with
        # TODO: All marker and line options could be defined as lists...
        n_labels = self.landmark_group.n_labels
        line_colour = _check_colours_list(
            render_lines, line_colour, n_labels,
            'Must pass a list of line colours with length n_labels or a single '
            'line colour for all labels.')
        marker_colour = _check_colours_list(
            render_markers, marker_colour, n_labels,
            'Must pass a list of marker colours with length n_labels or a '
            'single marker face colour for all labels.')
        marker_size = _parse_marker_size(marker_size, self.landmark_group.points)
        numbers_size = _parse_marker_size(numbers_size,
                                          self.landmark_group.points)

        # get pointcloud of each label
        sub_pointclouds = self._build_sub_pointclouds()

        # for each pointcloud
        # disabling the rendering greatly speeds up this for loop
        self.figure.scene.disable_render = True
        for i, (label, pc) in enumerate(sub_pointclouds):
            # render pointcloud
            pc.view(figure_id=self.figure_id, new_figure=False,
                    render_lines=render_lines, line_colour=line_colour[i],
                    line_width=line_width, render_markers=render_markers,
                    marker_style=marker_style, marker_size=marker_size,
                    marker_colour=marker_colour[i],
                    marker_resolution=marker_resolution, step=step,
                    alpha=alpha, render_numbering=render_numbering,
                    numbers_colour=numbers_colour, numbers_size=numbers_size)
        self.figure.scene.disable_render = False

        return self

    def _build_sub_pointclouds(self):
        return [(label, self.landmark_group.get_label(label))
                for label in self.landmark_group.labels]



def rasterize_mesh_from_barycentric_coordinate_images(mesh, bcoords_image,
                                                      tri_indices_image):
    r"""
    Renders an image of a `menpo.shape.TexturedTriMesh` or
    `menpo.shape.ColouredTriMesh` from a barycentric coordinate image pair.

    Note that the texture is rendered without any lighting model - think of
    this as a piecewise affine warp of the mesh's texture into the image (
    with z-buffering). As there is no lighting model, only meshes with
    colour/texture can be used with this method (a single color for the whole
    mesh would render flat with no shading).

    Parameters
    ----------
    mesh : `menpo.shape.TexturedTriMesh` or `menpo.shape.ColouredTriMesh`
        The 3D mesh who's texture will be rendered to the image.
    bcoords_image : `menpo.image.MaskedImage`
        The per-triangle barycentric coordinates for what should be rendered
        into each pixel. See :map:`rasterize_barycentric_coordinate_images`.
    tri_indices_image : `menpo.image.MaskedImage`
        The triangle index identifying the triangle that is visable at a pixel
        after z-buffering. See :map:`rasterize_barycentric_coordinate_images`.

    Returns
    -------
    `menpo.image.MaskedImage`
        A rasterized image of the mesh.
    """
    # Sample the mesh texture space to find the colors-per pixel
    colours = mesh.sample_texture_with_barycentric_coordinates(
        bcoords_image.as_vector(keep_channels=True).T,
        tri_indices_image.as_vector())
    # Rebuild the image using the usual from_vector machinery
    return tri_indices_image.from_vector(colours.T,
                                         n_channels=mesh.n_channels)


def rasterize_shape_image_from_barycentric_coordinate_images(mesh,
                                                             bcoords_image,
                                                             tri_indices_image):
    r"""
    Renders an XYZ shape image of a `menpo.shape.TexturedTriMesh` or
    `menpo.shape.ColouredTriMesh` from a barycentric coordinate image pair.


    Parameters
    ----------
    mesh : `menpo.shape.TexturedTriMesh` or `menpo.shape.ColouredTriMesh`
        The 3D mesh who's texture will be rendered to the image.
    bcoords_image : `menpo.image.MaskedImage`
        The per-triangle barycentric coordinates for what should be rendered
        into each pixel. See :map:`rasterize_barycentric_coordinate_images`.
    tri_indices_image : `menpo.image.MaskedImage`
        The triangle index identifying the triangle that is visable at a pixel
        after z-buffering. See :map:`rasterize_barycentric_coordinate_images`.

    Returns
    -------
    `menpo.image.MaskedImage`
        A rasterized shape image image of the mesh.
    """
    # Sample the mesh texture space to find the colors-per pixel
    shape_per_pixel = mesh.project_barycentric_coordinates(
        bcoords_image.as_vector(keep_channels=True).T,
        tri_indices_image.as_vector())
    # Rebuild the image using the usual from_vector machinery
    return tri_indices_image.from_vector(shape_per_pixel.points.T,
                                         n_channels=mesh.n_channels)


def rasterize_mesh(mesh_in_img, image_shape):
    from .cpu import rasterize_barycentric_coordinate_images
    bcs = rasterize_barycentric_coordinate_images(mesh_in_img, image_shape)
    return rasterize_mesh_from_barycentric_coordinate_images(mesh_in_img, *bcs)



import os
from pathlib import Path


def menpo3d_src_dir_path():
    r"""The path to the top of the menpo3d Python package.

    Useful for locating where the data folder is stored.

    Returns
    -------
    path : str
        The full path to the top of the Menpo3d package
    """
    return Path(os.path.abspath(__file__)).parent



import numpy as np

from menpo.io.output.base import _enforce_only_paths_supported
from menpo.shape.mesh import TexturedTriMesh


def obj_exporter(mesh, file_handle, **kwargs):
    r"""
    Given a file handle to write in to (which should act like a Python `file`
    object), write out the mesh data. No value is returned.

    Note that this does not save out textures of textured images, and so should
    not be used in isolation.

    Parameters
    ----------
    file_handle : `str`
        The full path where the obj will be saved out.
    mesh : :map:`TriMesh`
        Any subclass of :map:`TriMesh`. If :map:`TexturedTriMesh` texture
        coordinates will be saved out. Note that :map:`ColouredTriMesh`
        will only have shape data saved out, as .OBJ doesn't robustly support
        per-vertex colour information.
    """
    for v in mesh.points:
        file_handle.write('v {} {} {}\n'.format(*v).encode('utf-8'))
    file_handle.write(b'\n')
    if isinstance(mesh, TexturedTriMesh):
        for tc in mesh.tcoords.points:
            file_handle.write('vt {} {}\n'.format(*tc).encode('utf-8'))
        file_handle.write(b'\n')
        # triangulation of points and tcoords is identical
        for t in (mesh.trilist + 1):
            file_handle.write('f {0}/{0} {1}/{1} {2}/{2}\n'.format(*t).encode('utf-8'))
    else:
        # no tcoords - so triangulation is straight forward
        for t in (mesh.trilist + 1):
            file_handle.write('f {} {} {}\n'.format(*t).encode('utf-8'))


def ply_exporter(mesh, file_path, binary=False, **kwargs):
    r"""
    Given a file path to write in to write out the mesh data.
    No value is returned. Only file paths are supported and if a file handle
    is passed it will be ignored and a warning will be raised.

    Note that this does not save out textures of textured images, and so should
    not be used in isolation.

    Parameters
    ----------
    file_path : `str`
        The full path where the obj will be saved out.
    mesh : :map:`TriMesh`
        Any subclass of :map:`TriMesh`. If :map:`TexturedTriMesh` texture
        coordinates will be saved out. Note that :map:`ColouredTriMesh`
        will only have shape data saved out, as .PLY doesn't robustly support
        per-vertex colour information.
    binary: `bool`, optional
        Specify whether to format output in binary or ascii, defaults to False
    """
    import vtk
    from vtk.util.numpy_support import numpy_to_vtk, numpy_to_vtkIdTypeArray

    file_path = _enforce_only_paths_supported(file_path, 'PLY')

    polydata = vtk.vtkPolyData()
    points = vtk.vtkPoints()
    points.SetData(numpy_to_vtk(mesh.points))
    polydata.SetPoints(points)

    cells = vtk.vtkCellArray()
    counts = np.empty((mesh.trilist.shape[0], 1), dtype=np.int)
    counts.fill(3)
    tris = np.concatenate((counts, mesh.trilist), axis=1)
    cells.SetCells(mesh.trilist.shape[0], numpy_to_vtkIdTypeArray(tris))
    polydata.SetPolys(cells)

    if isinstance(mesh, TexturedTriMesh):
        pointdata = polydata.GetPointData()
        pointdata.SetTCoords(numpy_to_vtk(mesh.tcoords.points))

    ply_writer = vtk.vtkPLYWriter()
    ply_writer.SetFileName(str(file_path))
    ply_writer.SetInputData(polydata)
    if not binary:
        ply_writer.SetFileTypeToASCII()
    else:
        ply_writer.SetFileTypeToBinary()
    ply_writer.Update()
    ply_writer.Write()



import numpy as np
from collections import Iterable

from menpo.base import LazyList
from menpo.image import Image
from menpo.transform import Homogeneous
from menpo.shape import PointCloud

from menpo3d.extractimage import extract_per_vertex_colour
from menpo3d.rasterize import rasterize_mesh


# The below two functions are only temporarily in menpo3d. menpo will soon
# have a PR which adds these convenience methods on mesh subclasses.
#
# Once those are in we can remove these functions.

from menpo.base import copy_landmarks_and_path
from menpo.shape import TriMesh, ColouredTriMesh


def as_trimesh(self, copy=True):
    r"""
    Convert this :map:`TexturedTriMesh` to a :map:`TriMesh`, discarding the
    texture and tcoords.

    Parameters
    ----------
    copy : `bool`, optional
        If ``False``, the produced :map:`TriMesh` will share points and
        trilist with ``self``. Only suggested to be used for performance.

    Returns
    -------
    trimesh : :map:`TriMesh`
        A trimesh with the same points trilist and landmarks as this shape.
    """
    return copy_landmarks_and_path(self, TriMesh(self.points,
                                                 trilist=self.trilist,
                                                 copy=copy))


def as_colouredtrimesh(self, colours=None, copy=True):
    """
    Converts this to a :map:`ColouredTriMesh`.

    Parameters
    ----------
    colours : ``(N, 3)`` `ndarray`, optional
        The floating point RGB colour per vertex. If not given, grey will be
        assigned to each vertex.
    copy : `bool`, optional
        If ``True``, the graph will be a copy.

    Returns
    -------
    coloured : :map:`ColouredTriMesh`
        A version of this mesh with per-vertex colour assigned.
    """
    ctm = ColouredTriMesh(self.points, trilist=self.trilist,
                          colours=colours, copy=copy)
    return copy_landmarks_and_path(self, ctm)


def error_function(mesh, gt_mesh):
    return 1.


class Result(object):
    r"""
    Class for defining a basic fitting result. It holds the final mesh of a
    fitting process and, optionally, the initial mesh, ground truth mesh
    and the image object.

    Parameters
    ----------
    final_mesh : `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh`
        The final mesh of the fitting process.
    final_camera_transform : `menpo3d.camera.PerspectiveCamera`
        The final camera transform object.
    image : `menpo.image.Image` or `subclass` or ``None``, optional
        The image on which the fitting process was applied. Note that a copy
        of the image will be assigned as an attribute. If ``None``, then no
        image is assigned.
    initial_mesh : `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh` or ``None``, optional
        The initial mesh that was provided to the fitting method to
        initialise the fitting process. If ``None``, then no initial mesh is
        assigned.
    initial_camera_transform : `menpo3d.camera.PerspectiveCamera` or ``None``, optional
        The initial camera transform object. If ``None``, then no initial
        camera is assigned.
    gt_mesh : `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh` or ``None``, optional
        The ground truth mesh associated with the image. If ``None``, then no
        ground truth mesh is assigned.
    """
    def __init__(self, final_mesh, final_camera_transform, image=None,
                 initial_mesh=None, initial_camera_transform=None,
                 gt_mesh=None):
        self._final_mesh = final_mesh
        self._final_camera_transform = final_camera_transform
        self._initial_mesh = initial_mesh
        self._initial_camera_transform = initial_camera_transform
        self._gt_mesh = gt_mesh
        # If image is provided, create a copy
        self._image = None
        if image is not None:
            self._image = Image(image.pixels)

    @property
    def is_iterative(self):
        r"""
        Flag whether the object is an iterative fitting result.

        :type: `bool`
        """
        return False

    @property
    def final_mesh(self):
        r"""
        Returns the final mesh of the fitting process.

        :type: `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh`
        """
        return self._final_mesh

    @property
    def final_camera_transform(self):
        r"""
        Returns the final camera transform of the fitting process.

        :type: `menpo3d.camera.PerspectiveCamera`
        """
        return self._final_camera_transform

    @property
    def initial_mesh(self):
        r"""
        Returns the initial mesh that was provided to the fitting method to
        initialise the fitting process. In case the initial mesh does not
        exist, then ``None`` is returned.

        :type: `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh` or ``None``
        """
        return self._initial_mesh

    @property
    def initial_camera_transform(self):
        r"""
        Returns the initial camera transform of the fitting process.

        :type: `menpo3d.camera.PerspectiveCamera` or ``None``
        """
        return self._initial_camera_transform

    @property
    def gt_mesh(self):
        r"""
        Returns the ground truth mesh associated with the image. In case there
        is not an attached ground truth mesh, then ``None`` is returned.

        :type: `menpo.shape.TriMesh` or ``None``
        """
        return self._gt_mesh

    @property
    def image(self):
        r"""
        Returns the image that the fitting was applied on, if it was provided.
        Otherwise, it returns ``None``.

        :type: `menpo.shape.Image` or `subclass` or ``None``
        """
        return self._image

    def final_error(self, compute_error):
        r"""
        Returns the final error of the fitting process, if the ground truth
        mesh exists. This is the error computed based on the `final_mesh`.

        Parameters
        ----------
        compute_error: `callable` or ``None``, optional
            Callable that computes the error between the fitted and
            ground truth meshes.

        Returns
        -------
        final_error : `float`
            The final error at the end of the fitting process.

        Raises
        ------
        ValueError
            Ground truth mesh has not been set, so the final error cannot be
            computed
        """
        if compute_error is None:
            compute_error = error_function
        if self.gt_mesh is not None:
            return compute_error(self.final_mesh, self.gt_mesh)
        else:
            raise ValueError('Ground truth mesh has not been set, so the '
                             'final error cannot be computed')

    def initial_error(self, compute_error):
        r"""
        Returns the initial error of the fitting process, if the ground truth
        mesh and initial mesh exist. This is the error computed based on the
        `initial_mesh`.

        Parameters
        ----------
        compute_error: `callable` or ``None``, optional
            Callable that computes the error between the initial and
            ground truth meshes.

        Returns
        -------
        initial_error : `float`
            The initial error at the beginning of the fitting process.

        Raises
        ------
        ValueError
            Initial mesh has not been set, so the initial error cannot be
            computed
        ValueError
            Ground truth mesh has not been set, so the initial error cannot be
            computed
        """
        if compute_error is None:
            compute_error = error_function
        if self.initial_mesh is None:
            raise ValueError('Initial shape has not been set, so the initial '
                             'error cannot be computed')
        elif self.gt_mesh is None:
            raise ValueError('Ground truth shape has not been set, so the '
                             'initial error cannot be computed')
        else:
            return compute_error(self.initial_mesh, self.gt_mesh)

    def rasterized_final_mesh(self, shape=None):
        r"""
        Returns the rasterized final mesh. The image's shape will be used in
        case `shape` is not provided.

        Parameters
        ----------
        shape: `(int, int)` or ``None``, optional
            The shape of the rasterized image. If ``None``, then the fitted
            image's shape will be used.

        Returns
        -------
        rasterized_final_mesh : `menpo.image.Image`
            The image with the rasterized final mesh.

        Raises
        ------
        ValueError
            The final camera transform does not exist.
        ValueError
            You need to provide an image shape, since the image does not exist.
        """
        if self.final_camera_transform is None:
            raise ValueError("The final camera transform does not exist.")
        if shape is None:
            if self.image is None:
                raise ValueError("You need to provide an image shape, "
                                 "since the image does not exist.")
            else:
                shape = self.image.shape
        return rasterize_mesh(
            self.final_camera_transform.apply(self.final_mesh), shape)

    def rasterized_initial_mesh(self, shape=None):
        r"""
        Returns the rasterized initial mesh, if it exists. The image's shape
        will be used in case `shape` is not provided.

        Parameters
        ----------
        shape: `(int, int)` or ``None``, optional
            The shape of the rasterized image. If ``None``, then the fitted
            image's shape will be used.

        Returns
        -------
        rasterized_initial_mesh : `menpo.image.Image`
            The image with the rasterized initial mesh.

        Raises
        ------
        ValueError
            The initial mesh does not exist.
        ValueError
            The initial camera transform does not exist.
        ValueError
            You need to provide an image shape, since the image does not exist.
        """
        if self.initial_mesh is None:
            raise ValueError("The initial mesh does not exist.")
        if self.initial_camera_transform is None:
            raise ValueError("The initial camera transform does not exist.")
        if shape is None:
            if self.image is None:
                raise ValueError("You need to provide an image shape, "
                                 "since the image does not exist.")
            else:
                shape = self.image.shape
        return rasterize_mesh(
            self.initial_camera_transform.apply(self.initial_mesh), shape)

    def view_final_mesh(self, figure_id=None, new_figure=False,
                        textured=True, mesh_type='surface',
                        mesh_colour=(1, 0, 0), line_width=2,
                        ambient_light=0.0, specular_light=0.0, step=None,
                        alpha=1.0):
        """
        Visualize the final mesh.

        Parameters
        ----------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        textured : `bool`, optional
            If ``True``, then the texture is rendered. If ``False``, then only
            the TriMesh is rendered with the specified `colour`.
        mesh_type : ``{'surface', 'wireframe'}``, optional
            The representation type to be used for the mesh.
        mesh_colour : `(float, float, float)`, optional
            The colour of the mesh as a tuple of RGB values. It only applies if
            `textured` is ``False``.
        line_width : `float`, optional
            The width of the lines, if there are any.
        ambient_light : `float`, optional
            The ambient light intensity. It must be in range ``[0., 1.]``.
        specular_light : `float`, optional
            The specular light intensity. It must be in range ``[0., 1.]``.
        step : `int` or ``None``, optional
            If `int`, then one every `step` normals will be rendered.
            If ``None``, then all vertexes will be rendered. It only applies if
            `normals` is not ``None``.
        alpha : `float`, optional
            Defines the transparency (opacity) of the object.

        Returns
        -------
        renderer : `menpo3d.visualize.TexturedTriMeshViewer3D`
            The Menpo3D rendering object.
        """
        return self.final_mesh.view(
            figure_id=figure_id, new_figure=new_figure, textured=textured,
            mesh_type=mesh_type, mesh_colour=mesh_colour,
            line_width=line_width, ambient_light=ambient_light,
            specular_light=specular_light, step=step, alpha=alpha)

    def view_initial_mesh(self, figure_id=None, new_figure=False,
                          textured=True, mesh_type='surface',
                          mesh_colour=(1, 0, 0), line_width=2,
                          ambient_light=0.0, specular_light=0.0, step=None,
                          alpha=1.0):
        """
        Visualize the initial mesh, if it exists.

        Parameters
        ----------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        textured : `bool`, optional
            If ``True``, then the texture is rendered. If ``False``, then only
            the TriMesh is rendered with the specified `colour`.
        mesh_type : ``{'surface', 'wireframe'}``, optional
            The representation type to be used for the mesh.
        mesh_colour : `(float, float, float)`, optional
            The colour of the mesh as a tuple of RGB values. It only applies if
            `textured` is ``False``.
        line_width : `float`, optional
            The width of the lines, if there are any.
        ambient_light : `float`, optional
            The ambient light intensity. It must be in range ``[0., 1.]``.
        specular_light : `float`, optional
            The specular light intensity. It must be in range ``[0., 1.]``.
        step : `int` or ``None``, optional
            If `int`, then one every `step` normals will be rendered.
            If ``None``, then all vertexes will be rendered. It only applies if
            `normals` is not ``None``.
        alpha : `float`, optional
            Defines the transparency (opacity) of the object.

        Returns
        -------
        renderer : `menpo3d.visualize.TexturedTriMeshViewer3D`
            The Menpo3D rendering object.
        """
        if self.initial_mesh is None:
            raise ValueError("The initial mesh does not exist.")
        else:
            return self.initial_mesh.view(
                figure_id=figure_id, new_figure=new_figure, textured=textured,
                mesh_type=mesh_type, mesh_colour=mesh_colour,
                line_width=line_width, ambient_light=ambient_light,
                specular_light=specular_light, step=step, alpha=alpha)

    def view_gt_mesh(self, figure_id=None, new_figure=False,
                     mesh_type='wireframe', line_width=2, colour=(1, 0, 0),
                     marker_style='sphere', marker_size=0.05,
                     marker_resolution=8, step=None, alpha=1.0):
        """
        Visualize the ground truth mesh, if it exists.

        Parameters
        ----------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        mesh_type : `str`, optional
            The representation type to be used for the mesh.
            Example options ::

                {surface, wireframe, points, mesh, fancymesh}

        line_width : `float`, optional
            The width of the lines, if there are any.
        colour : `(float, float, float)`, optional
            The colour of the mesh as a tuple of RGB values.
        marker_style : `str`, optional
            The style of the markers.
            Example options ::

                {2darrow, 2dcircle, 2dcross, 2ddash, 2ddiamond, 2dhooked_arrow,
                 2dsquare, 2dthick_arrow, 2dthick_cross, 2dtriangle, 2dvertex,
                 arrow, axes, cone, cube, cylinder, point, sphere}

        marker_size : `float`, optional
            The size of the markers. This size can be seen as a scale factor
            applied to the size markers, which is by default calculated from
            the inter-marker spacing. It only applies for the 'fancymesh'.
        marker_resolution : `int`, optional
            The resolution of the markers. For spheres, for instance, this is
            the number of divisions along theta and phi. It only applies for
            the 'fancymesh'.
        step : `int` or ``None``, optional
            If `int`, then one every `step` markers will be rendered.
            If ``None``, then all vertexes will be rendered. It only applies for
            the 'fancymesh' and if `normals` is not ``None``.
        alpha : `float`, optional
            Defines the transparency (opacity) of the object.

        Returns
        -------
        renderer : `menpo3d.visualize.TriMeshViewer3D`
            The Menpo3D rendering object.
        """
        if self.initial_mesh is None:
            raise ValueError("The ground truth mesh does not exist.")
        else:
            return self.gt_mesh.view(
                figure_id=figure_id, new_figure=new_figure, mesh_type=mesh_type,
                line_width=line_width, colour=colour, marker_style=marker_style,
                marker_size=marker_size, marker_resolution=marker_resolution,
                step=step, alpha=alpha)

    def __str__(self):
        out = "Fitting result of mesh with {} points.".format(
            self.final_mesh.n_points)
        if self.gt_mesh is not None:
            if self.initial_mesh is not None:
                out += "\nInitial error: {:.4f}".format(self.initial_error())
            out += "\nFinal error: {:.4f}".format(self.final_error())
        return out


class NonParametricIterativeResult(Result):
    r"""
    Class for defining a non-parametric iterative fitting result, i.e. the
    result of a method that does not optimize over a parametric shape model.
    It holds the meshes of all the iterations of the fitting procedure.
    It can optionally store the image on which the fitting was applied,
    as well as its ground truth mesh.

    Parameters
    ----------
    meshes : `list` of `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh`
        The `list` of meshes per iteration. Note that the list does not
        include the initial mesh. However, it includes the reconstruction of
        the initial mesh. The last member of the list is the final mesh.
    camera_transforms : `list` of `menpo3d.camera.PerspectiveCamera`
        The `list` of camera transform objects per iteration. Note that the
        list does not include the initial camera transform. The last member
        of the list is the final camera transform.
    image : `menpo.image.Image` or `subclass` or ``None``, optional
        The image on which the fitting process was applied. Note that a copy
        of the image will be assigned as an attribute. If ``None``, then no
        image is assigned.
    initial_mesh : `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh` or ``None``, optional
        The initial mesh from which the fitting process started. If
        ``None``, then no initial mesh is assigned.
    initial_camera_transform : `menpo3d.camera.PerspectiveCamera` or ``None``, optional
        The initial camera transform. If ``None``, then no initial camera
        is assigned.
    gt_mesh : `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh` or ``None``, optional
        The ground truth mesh associated with the image. If ``None``, then
        no ground truth mesh is assigned.
    costs : `list` of `float` or ``None``, optional
        The `list` of cost per iteration. If ``None``, then it is assumed
        that the cost function cannot be computed for the specific
        algorithm. It must have the same length as `meshes`.
    """
    def __init__(self, meshes, camera_transforms, image=None, initial_mesh=None,
                 initial_camera_transform=None, gt_mesh=None, costs=None):
        super(NonParametricIterativeResult, self).__init__(
            final_mesh=meshes[-1], final_camera_transform=camera_transforms[-1],
            image=image, initial_mesh=initial_mesh,
            initial_camera_transform=initial_camera_transform, gt_mesh=gt_mesh)
        self._n_iters = len(meshes)
        # If initial mesh is provided, then add it in the beginning of meshes
        self._meshes = meshes
        if self.initial_mesh is not None:
            self._meshes = [self.initial_mesh] + self._meshes
        # If initial camera transform is provided, then add it in the beginning
        # of meshes
        self._camera_transforms = camera_transforms
        if self.initial_camera_transform is not None:
            self._camera_transforms = ([self.initial_camera_transform] +
                                       self._camera_transforms)
        # Add costs as property
        self._costs = costs

    @property
    def is_iterative(self):
        r"""
        Flag whether the object is an iterative fitting result.

        :type: `bool`
        """
        return True

    @property
    def meshes(self):
        r"""
        Returns the `list` of meshes obtained at each iteration of the fitting
        process. The `list` includes the `initial_mesh` (if it exists) and
        `final_mesh`.

        :type: `list` of `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh`
        """
        return self._meshes

    @property
    def camera_transforms(self):
        r"""
        Returns the `list` of camera transforms per iteration of the fitting
        process. The `list` includes the `initial_camera_transform` (if it
        exists) and the camera transform of the final mesh.

        :type: `list` of `menpo3d.camera.PerspectiveCamera`
        """
        return self._camera_transforms

    @property
    def n_iters(self):
        r"""
        Returns the total number of iterations of the fitting process.

        :type: `int`
        """
        return self._n_iters

    def to_result(self, pass_image=True, pass_initial_mesh=True,
                  pass_gt_mesh=True):
        r"""
        Returns a :map:`Result` instance of the object, i.e. a fitting result
        object that does not store the iterations. This can be useful for
        reducing the size of saved fitting results.

        Parameters
        ----------
        pass_image : `bool`, optional
            If ``True``, then the image will get passed (if it exists).
        pass_initial_mesh : `bool`, optional
            If ``True``, then the initial mesh will get passed (if it exists).
        pass_gt_mesh : `bool`, optional
            If ``True``, then the ground truth mesh will get passed (if it
            exists).

        Returns
        -------
        result : :map:`Result`
            The final "lightweight" fitting result.
        """
        image = None
        if pass_image:
            image = self.image
        initial_mesh = None
        initial_camera_transform = None
        if pass_initial_mesh:
            initial_mesh = self.initial_mesh
            initial_camera_transform = self.initial_camera_transform
        gt_mesh = None
        if pass_gt_mesh:
            gt_mesh = self.gt_mesh
        return Result(self.final_mesh, self.final_camera_transform,
                      image=image, initial_mesh=initial_mesh,
                      initial_camera_transform=initial_camera_transform,
                      gt_mesh=gt_mesh)

    def errors(self, compute_error=None):
        r"""
        Returns a list containing the error at each fitting iteration, if the
        ground truth mesh exists.

        Parameters
        ----------
        compute_error: `callable`, optional
            Callable that computes the error between the mesh at each
            iteration and the ground truth mesh.

        Returns
        -------
        errors : `list` of `float`
            The error at each iteration of the fitting process.

        Raises
        ------
        ValueError
            Ground truth mesh has not been set, so the final error cannot be
            computed
        """
        if compute_error is None:
            compute_error = error_function
        if self.gt_mesh is not None:
            return [compute_error(t, self.gt_mesh)
                    for t in self.meshes]
        else:
            raise ValueError('Ground truth shape has not been set, so the '
                             'errors per iteration cannot be computed')

    def plot_errors(self, compute_error=None, figure_id=None,
                    new_figure=False, render_lines=True, line_colour='b',
                    line_style='-', line_width=2, render_markers=True,
                    marker_style='o', marker_size=4, marker_face_colour='b',
                    marker_edge_colour='k', marker_edge_width=1.,
                    render_axes=True, axes_font_name='sans-serif',
                    axes_font_size=10, axes_font_style='normal',
                    axes_font_weight='normal', axes_x_limits=0.,
                    axes_y_limits=None, axes_x_ticks=None,
                    axes_y_ticks=None, figure_size=(10, 6),
                    render_grid=True, grid_line_style='--',
                    grid_line_width=0.5):
        r"""
        Plot of the error evolution at each fitting iteration.

        Parameters
        ----------
        compute_error: `callable`, optional
            Callable that computes the error between the shape at each
            iteration and the ground truth shape.
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        render_lines : `bool`, optional
            If ``True``, the line will be rendered.
        line_colour : `colour` or ``None`` (See below), optional
            The colour of the line. If ``None``, the colour is sampled from
            the jet colormap.
            Example `colour` options are ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        line_style : `str` (See below), optional
            The style of the lines. Example options::

                {-, --, -., :}

        line_width : `float`, optional
            The width of the lines.
        render_markers : `bool`, optional
            If ``True``, the markers will be rendered.
        marker_style : `str` (See below), optional
            The style of the markers.
            Example `marker` options ::

                {., ,, o, v, ^, <, >, +, x, D, d, s, p, *, h, H, 1, 2, 3, 4, 8}

        marker_size : `int`, optional
            The size of the markers in points.
        marker_face_colour : `colour` or ``None``, optional
            The face (filling) colour of the markers. If ``None``, the colour
            is sampled from the jet colormap.
            Example `colour` options are ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_colour : `colour` or ``None``, optional
            The edge colour of the markers. If ``None``, the colour
            is sampled from the jet colormap.
            Example `colour` options are ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_width : `float`, optional
            The width of the markers' edge.
        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : `str` (See below), optional
            The font of the axes.
            Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : `str` (See below), optional
            The font style of the axes.
            Example options ::

                {normal, italic, oblique}

        axes_font_weight : `str` (See below), optional
            The font weight of the axes.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                 semibold, demibold, demi, bold, heavy, extra bold, black}

        axes_x_limits : `float` or (`float`, `float`) or ``None``, optional
            The limits of the x axis. If `float`, then it sets padding on the
            right and left of the graph as a percentage of the curves' width. If
            `tuple` or `list`, then it defines the axis limits. If ``None``,
            then the limits are set automatically.
        axes_y_limits : `float` or (`float`, `float`) or ``None``, optional
            The limits of the y axis. If `float`, then it sets padding on the
            top and bottom of the graph as a percentage of the curves' height.
            If `tuple` or `list`, then it defines the axis limits. If ``None``,
            then the limits are set automatically.
        axes_x_ticks : `list` or `tuple` or ``None``, optional
            The ticks of the x axis.
        axes_y_ticks : `list` or `tuple` or ``None``, optional
            The ticks of the y axis.
        figure_size : (`float`, `float`) or ``None``, optional
            The size of the figure in inches.
        render_grid : `bool`, optional
            If ``True``, the grid will be rendered.
        grid_line_style : ``{'-', '--', '-.', ':'}``, optional
            The style of the grid lines.
        grid_line_width : `float`, optional
            The width of the grid lines.

        Returns
        -------
        renderer : `menpo.visualize.GraphPlotter`
            The renderer object.
        """
        from menpo.visualize import plot_curve
        errors = self.errors(compute_error=compute_error)
        return plot_curve(
            x_axis=list(range(len(errors))), y_axis=[errors], figure_id=figure_id,
            new_figure=new_figure, title='Fitting Errors per Iteration',
            x_label='Iteration', y_label='Fitting Error',
            axes_x_limits=axes_x_limits, axes_y_limits=axes_y_limits,
            axes_x_ticks=axes_x_ticks, axes_y_ticks=axes_y_ticks,
            render_lines=render_lines, line_colour=line_colour,
            line_style=line_style, line_width=line_width,
            render_markers=render_markers, marker_style=marker_style,
            marker_size=marker_size, marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width, render_legend=False,
            render_axes=render_axes, axes_font_name=axes_font_name,
            axes_font_size=axes_font_size, axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight, figure_size=figure_size,
            render_grid=render_grid,  grid_line_style=grid_line_style,
            grid_line_width=grid_line_width)

    def displacements(self):
        r"""
        A list containing the displacement between the mesh of each iteration
        and the mesh of the previous one.

        :type: `list` of `ndarray`
        """
        return [np.linalg.norm(s1.points - s2.points, axis=1)
                for s1, s2 in zip(self.meshes, self.meshes[1:])]

    def displacements_stats(self, stat_type='mean'):
        r"""
        A list containing a statistical metric on the displacements between
        the mesh of each iteration and the mesh of the previous one.

        Parameters
        ----------
        stat_type : ``{'mean', 'median', 'min', 'max'}``, optional
            Specifies a statistic metric to be extracted from the displacements.

        Returns
        -------
        displacements_stat : `list` of `float`
            The statistical metric on the points displacements for each
            iteration.

        Raises
        ------
        ValueError
            type must be 'mean', 'median', 'min' or 'max'
        """
        if stat_type == 'mean':
            return [np.mean(d) for d in self.displacements()]
        elif stat_type == 'median':
            return [np.median(d) for d in self.displacements()]
        elif stat_type == 'max':
            return [np.max(d) for d in self.displacements()]
        elif stat_type == 'min':
            return [np.min(d) for d in self.displacements()]
        else:
            raise ValueError("type must be 'mean', 'median', 'min' or 'max'")

    def plot_displacements(self, stat_type='mean', figure_id=None,
                           new_figure=False, render_lines=True, line_colour='b',
                           line_style='-', line_width=2, render_markers=True,
                           marker_style='o', marker_size=4,
                           marker_face_colour='b', marker_edge_colour='k',
                           marker_edge_width=1., render_axes=True,
                           axes_font_name='sans-serif', axes_font_size=10,
                           axes_font_style='normal', axes_font_weight='normal',
                           axes_x_limits=0., axes_y_limits=None,
                           axes_x_ticks=None, axes_y_ticks=None,
                           figure_size=(10, 6), render_grid=True,
                           grid_line_style='--', grid_line_width=0.5):
        r"""
        Plot of a statistical metric of the displacement between the mesh of
        each iteration and the mesh of the previous one.

        Parameters
        ----------
        stat_type : {``mean``, ``median``, ``min``, ``max``}, optional
            Specifies a statistic metric to be extracted from the displacements
            (see also `displacements_stats()` method).
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        render_lines : `bool`, optional
            If ``True``, the line will be rendered.
        line_colour : `colour` or ``None`` (See below), optional
            The colour of the line. If ``None``, the colour is sampled from
            the jet colormap.
            Example `colour` options are ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        line_style : `str` (See below), optional
            The style of the lines. Example options::

                {-, --, -., :}

        line_width : `float`, optional
            The width of the lines.
        render_markers : `bool`, optional
            If ``True``, the markers will be rendered.
        marker_style : `str` (See below), optional
            The style of the markers.
            Example `marker` options ::

                {., ,, o, v, ^, <, >, +, x, D, d, s, p, *, h, H, 1, 2, 3, 4, 8}

        marker_size : `int`, optional
            The size of the markers in points.
        marker_face_colour : `colour` or ``None``, optional
            The face (filling) colour of the markers. If ``None``, the colour
            is sampled from the jet colormap.
            Example `colour` options are ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_colour : `colour` or ``None``, optional
            The edge colour of the markers. If ``None``, the colour
            is sampled from the jet colormap.
            Example `colour` options are ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_width : `float`, optional
            The width of the markers' edge.
        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : `str` (See below), optional
            The font of the axes.
            Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : `str` (See below), optional
            The font style of the axes.
            Example options ::

                {normal, italic, oblique}

        axes_font_weight : `str` (See below), optional
            The font weight of the axes.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                 semibold, demibold, demi, bold, heavy, extra bold, black}

        axes_x_limits : `float` or (`float`, `float`) or ``None``, optional
            The limits of the x axis. If `float`, then it sets padding on the
            right and left of the graph as a percentage of the curves' width. If
            `tuple` or `list`, then it defines the axis limits. If ``None``,
            then the limits are set automatically.
        axes_y_limits : `float` or (`float`, `float`) or ``None``, optional
            The limits of the y axis. If `float`, then it sets padding on the
            top and bottom of the graph as a percentage of the curves' height.
            If `tuple` or `list`, then it defines the axis limits. If ``None``,
            then the limits are set automatically.
        axes_x_ticks : `list` or `tuple` or ``None``, optional
            The ticks of the x axis.
        axes_y_ticks : `list` or `tuple` or ``None``, optional
            The ticks of the y axis.
        figure_size : (`float`, `float`) or ``None``, optional
            The size of the figure in inches.
        render_grid : `bool`, optional
            If ``True``, the grid will be rendered.
        grid_line_style : ``{'-', '--', '-.', ':'}``, optional
            The style of the grid lines.
        grid_line_width : `float`, optional
            The width of the grid lines.

        Returns
        -------
        renderer : `menpo.visualize.GraphPlotter`
            The renderer object.
        """
        from menpo.visualize import plot_curve
        # set labels
        if stat_type == 'max':
            name = 'Maximum'
        elif stat_type == 'min':
            name = 'Minimum'
        elif stat_type == 'mean':
            name = 'Mean'
        elif stat_type == 'median':
            name = 'Median'
        else:
            raise ValueError('stat_type must be one of {max, min, mean, '
                             'median}.')
        y_label = '{} Displacement'.format(name)
        title = '{} displacement per Iteration'.format(name)

        # plot
        displacements = self.displacements_stats(stat_type=stat_type)
        return plot_curve(
            x_axis=list(range(len(displacements))), y_axis=[displacements],
            figure_id=figure_id, new_figure=new_figure, title=title,
            x_label='Iteration', y_label=y_label,
            axes_x_limits=axes_x_limits, axes_y_limits=axes_y_limits,
            axes_x_ticks=axes_x_ticks, axes_y_ticks=axes_y_ticks,
            render_lines=render_lines, line_colour=line_colour,
            line_style=line_style, line_width=line_width,
            render_markers=render_markers, marker_style=marker_style,
            marker_size=marker_size, marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width, render_legend=False,
            render_axes=render_axes, axes_font_name=axes_font_name,
            axes_font_size=axes_font_size, axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight, figure_size=figure_size,
            render_grid=render_grid,  grid_line_style=grid_line_style,
            grid_line_width=grid_line_width)

    @property
    def costs(self):
        r"""
        Returns a `list` with the cost per iteration. It returns ``None`` if
        the costs are not computed.

        :type: `list` of `float` or ``None``
        """
        return self._costs

    def plot_costs(self, figure_id=None, new_figure=False, render_lines=True,
                   line_colour='b', line_style='-', line_width=2,
                   render_markers=True, marker_style='o', marker_size=4,
                   marker_face_colour='b', marker_edge_colour='k',
                   marker_edge_width=1., render_axes=True,
                   axes_font_name='sans-serif', axes_font_size=10,
                   axes_font_style='normal', axes_font_weight='normal',
                   axes_x_limits=0., axes_y_limits=None, axes_x_ticks=None,
                   axes_y_ticks=None, figure_size=(10, 6),
                   render_grid=True, grid_line_style='--',
                   grid_line_width=0.5):
        r"""
        Plot of the cost function evolution at each fitting iteration.

        Parameters
        ----------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        render_lines : `bool`, optional
            If ``True``, the line will be rendered.
        line_colour : `colour` or ``None``, optional
            The colour of the line. If ``None``, the colour is sampled from
            the jet colormap.
            Example `colour` options are ::

                    {'r', 'g', 'b', 'c', 'm', 'k', 'w'}
                    or
                    (3, ) ndarray

        line_style : ``{'-', '--', '-.', ':'}``, optional
            The style of the lines.
        line_width : `float`, optional
            The width of the lines.
        render_markers : `bool`, optional
            If ``True``, the markers will be rendered.
        marker_style : `marker`, optional
            The style of the markers.
            Example `marker` options ::

                    {'.', ',', 'o', 'v', '^', '<', '>', '+', 'x', 'D', 'd', 's',
                     'p', '*', 'h', 'H', '1', '2', '3', '4', '8'}

        marker_size : `int`, optional
            The size of the markers in points.
        marker_face_colour : `colour` or ``None``, optional
            The face (filling) colour of the markers. If ``None``, the colour
            is sampled from the jet colormap.
            Example `colour` options are ::

                    {'r', 'g', 'b', 'c', 'm', 'k', 'w'}
                    or
                    (3, ) ndarray

        marker_edge_colour : `colour` or ``None``, optional
            The edge colour of the markers.If ``None``, the colour
            is sampled from the jet colormap.
            Example `colour` options are ::

                    {'r', 'g', 'b', 'c', 'm', 'k', 'w'}
                    or
                    (3, ) ndarray

        marker_edge_width : `float`, optional
            The width of the markers' edge.
        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : See below, optional
            The font of the axes.
            Example options ::

                {'serif', 'sans-serif', 'cursive', 'fantasy', 'monospace'}

        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : ``{'normal', 'italic', 'oblique'}``, optional
            The font style of the axes.
        axes_font_weight : See below, optional
            The font weight of the axes.
            Example options ::

                {'ultralight', 'light', 'normal', 'regular', 'book', 'medium',
                 'roman', 'semibold', 'demibold', 'demi', 'bold', 'heavy',
                 'extra bold', 'black'}

        axes_x_limits : `float` or (`float`, `float`) or ``None``, optional
            The limits of the x axis. If `float`, then it sets padding on the
            right and left of the graph as a percentage of the curves' width. If
            `tuple` or `list`, then it defines the axis limits. If ``None``,
            then the limits are set automatically.
        axes_y_limits : `float` or (`float`, `float`) or ``None``, optional
            The limits of the y axis. If `float`, then it sets padding on the
            top and bottom of the graph as a percentage of the curves' height.
            If `tuple` or `list`, then it defines the axis limits. If ``None``,
            then the limits are set automatically.
        axes_x_ticks : `list` or `tuple` or ``None``, optional
            The ticks of the x axis.
        axes_y_ticks : `list` or `tuple` or ``None``, optional
            The ticks of the y axis.
        figure_size : (`float`, `float`) or ``None``, optional
            The size of the figure in inches.
        render_grid : `bool`, optional
            If ``True``, the grid will be rendered.
        grid_line_style : ``{'-', '--', '-.', ':'}``, optional
            The style of the grid lines.
        grid_line_width : `float`, optional
            The width of the grid lines.

        Returns
        -------
        renderer : `menpo.visualize.GraphPlotter`
            The renderer object.
        """
        from menpo.visualize import plot_curve
        costs = self.costs
        if costs is not None:
            return plot_curve(
                x_axis=list(range(len(costs))), y_axis=[costs],
                figure_id=figure_id, new_figure=new_figure,
                title='Cost per Iteration', x_label='Iteration',
                y_label='Cost Function', axes_x_limits=axes_x_limits,
                axes_y_limits=axes_y_limits, axes_x_ticks=axes_x_ticks,
                axes_y_ticks=axes_y_ticks, render_lines=render_lines,
                line_colour=line_colour, line_style=line_style,
                line_width=line_width, render_markers=render_markers,
                marker_style=marker_style, marker_size=marker_size,
                marker_face_colour=marker_face_colour,
                marker_edge_colour=marker_edge_colour,
                marker_edge_width=marker_edge_width, render_legend=False,
                render_axes=render_axes, axes_font_name=axes_font_name,
                axes_font_size=axes_font_size,
                axes_font_style=axes_font_style,
                axes_font_weight=axes_font_weight, figure_size=figure_size,
                render_grid=render_grid,  grid_line_style=grid_line_style,
                grid_line_width=grid_line_width)
        else:
            raise ValueError('costs are either not returned or not well '
                             'defined for the selected fitting algorithm')


class ParametricIterativeResult(NonParametricIterativeResult):
    r"""
    Class for defining a parametric iterative fitting result, i.e. the
    result of a method that optimizes the parameters of a shape model. It holds
    the shape parameters and camera transforms of all the iterations of the
    fitting procedure. It can optionally store the meshes, image on which the
    fitting was applied, as well as its ground truth mesh.

    .. note:: When using a method with a parametric shape model, the first step
              is to **reconstruct the initial mesh** using the shape model. The
              generated reconstructed mesh is then used as initialisation for
              the iterative optimisation. This step is not counted in the number
              of iterations.

    Parameters
    ----------
    shape_parameters : `list` of `ndarray`
        The `list` of shape parameters per iteration. Note that the list
        includes the parameters of the projection of the initial mesh. The last
        member of the list corresponds to the final mesh's parameters. It must
        have the same length as `camera_transforms` and `meshes`.
    meshes : `list` of `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh`
        The `list` of meshes per iteration. Note that the list does not
        include the initial mesh. However, it includes the reconstruction of
        the initial mesh. The last member of the list is the final mesh.
    camera_transforms : `list` of `menpo3d.camera.PerspectiveCamera`
        The `list` of camera transform objects per iteration. Note that the
        list does not include the initial camera transform. The last member
        of the list is the final camera transform.
    image : `menpo.image.Image` or `subclass` or ``None``, optional
        The image on which the fitting process was applied. Note that a copy
        of the image will be assigned as an attribute. If ``None``, then no
        image is assigned.
    initial_mesh : `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh` or ``None``, optional
        The initial mesh from which the fitting process started. If
        ``None``, then no initial mesh is assigned.
    initial_camera_transform : `menpo3d.camera.PerspectiveCamera` or ``None``, optional
        The initial camera transform. If ``None``, then no initial camera
        is assigned.
    gt_mesh : `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh` or ``None``, optional
        The ground truth mesh associated with the image. If ``None``, then
        no ground truth mesh is assigned.
    costs : `list` of `float` or ``None``, optional
        The `list` of cost per iteration. If ``None``, then it is assumed
        that the cost function cannot be computed for the specific
        algorithm. It must have the same length as `meshes`.
    """
    def __init__(self, shape_parameters, meshes, camera_transforms,
                 image=None, initial_mesh=None,
                 initial_camera_transform=None, gt_mesh=None, costs=None):
        # Assign shape parameters
        self._shape_parameters = shape_parameters
        # Get reconstructed initial shape
        self._reconstructed_initial_mesh = meshes[0]
        # Call superclass
        super(ParametricIterativeResult, self).__init__(
            meshes=meshes, camera_transforms=camera_transforms, image=image,
            initial_mesh=initial_mesh,
            initial_camera_transform=initial_camera_transform,
            gt_mesh=gt_mesh, costs=costs)
        # Correct n_iters. The initial mesh's reconstruction should not count
        # in the number of iterations.
        self._n_iters -= 1

    @property
    def meshes(self):
        r"""
        Returns the `list` of meshes obtained at each iteration of the fitting
        process. The `list` includes the `initial_mesh` (if it exists),
        `reconstructed_initial_mesh` and `final_mesh`.

        :type: `list` of `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh`
        """
        return self._meshes

    @property
    def camera_transforms(self):
        r"""
        Returns the `list` of camera transforms per iteration of the fitting
        process. The `list` includes the `initial_camera_transform` (if it
        exists) and the camera transform of the reconstructed initial mesh and
        final mesh.

        :type: `list` of `menpo3d.camera.PerspectiveCamera`
        """
        return self._camera_transforms

    @property
    def shape_parameters(self):
        r"""
        Returns the `list` of shape parameters obtained at each iteration of
        the fitting process. The `list` includes the parameters of the
        `reconstructed_initial_mesh` and `final_mesh`.

        :type: `list` of ``(n_params,)`` `ndarray`
        """
        return self._shape_parameters

    @property
    def reconstructed_initial_mesh(self):
        r"""
        Returns the initial mesh's reconstruction with the shape model that was
        used to initialise the iterative optimisation process.

        :type: `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh`
        """
        if self.initial_mesh is not None:
            return self.meshes[1]
        else:
            return self.meshes[0]

    @property
    def reconstructed_initial_camera_transform(self):
        r"""
        Returns the camera transform of the reconstructed initial mesh.

        :type: `menpo3d.camera.PerspectiveCamera`
        """
        if self.initial_camera_transform is not None:
            return self.camera_transforms[1]
        else:
            return self.camera_transforms[0]

    def rasterized_reconstructed_initial_mesh(self, shape=None):
        r"""
        Returns the rasterized reconstructed initial mesh, if it exists. The
        image's shape will be used in case `shape` is not provided.

        Parameters
        ----------
        shape: `(int, int)` or ``None``, optional
            The shape of the rasterized image. If ``None``, then the fitted
            image's shape will be used.

        Returns
        -------
        rasterized_reconstructed_initial_mesh : `menpo.image.Image`
            The image with the rasterized reconstructed initial mesh.

        Raises
        ------
        ValueError
            You need to provide an image shape, since the image does not exist.
        """
        if shape is None:
            if self.image is None:
                raise ValueError("You need to provide an image shape, "
                                 "since the image does not exist.")
            else:
                shape = self.image.shape
        return rasterize_mesh(
            self.reconstructed_initial_camera_transform.apply(
                self.reconstructed_initial_mesh), shape)

    @property
    def _reconstruction_indices(self):
        r"""
        Returns a list with the indices of reconstructed meshes in the `meshes`
        list.

        :type: `list` of `int`
        """
        if self.initial_mesh is not None:
            return [1]
        else:
            return [0]

    def reconstructed_initial_error(self, compute_error=None):
        r"""
        Returns the error of the reconstructed initial mesh of the fitting
        process, if the ground truth mesh exists. This is the error computed
        based on the `reconstructed_initial_mesh`.

        Parameters
        ----------
        compute_error: `callable`, optional
            Callable that computes the error between the reconstructed initial
            and ground truth meshes.

        Returns
        -------
        reconstructed_initial_error : `float`
            The error that corresponds to the initial mesh's reconstruction.

        Raises
        ------
        ValueError
            Ground truth mesh has not been set, so the reconstructed initial
            error cannot be computed
        """
        if compute_error is None:
            compute_error = error_function
        if self.gt_mesh is None:
            raise ValueError('Ground truth mesh has not been set, so the '
                             'reconstructed initial error cannot be computed')
        else:
            return compute_error(self.reconstructed_initial_mesh, self.gt_mesh)


def _affine_2d_to_3d(transform):
    h_matrix = np.eye(4)
    h_matrix[:2, :2] = transform.h_matrix[:2, :2]
    h_matrix[:2, 3] = transform.h_matrix[:2, 2]
    h_matrix[3, :2] = transform.h_matrix[2, :2]
    return Homogeneous(h_matrix)


class MultiScaleNonParametricIterativeResult(NonParametricIterativeResult):
    r"""
    Class for defining a multi-scale non-parametric iterative fitting result,
    i.e. the result of a multi-scale method that does not optimise over a
    parametric shape model. It holds the meshes of all the iterations of
    the fitting procedure, as well as the scales. It can optionally store the
    image on which the fitting was applied, as well as its ground truth mesh.

    Parameters
    ----------
    results : `list` of :map:`NonParametricIterativeResult`
        The `list` of non parametric iterative results per scale.
    n_scales : `int`
        The number of scales.
    affine_transforms : `list` of `menpo.transform.Affine`
        The list of affine transforms per scale that transform where used
        dutring the image's pre-processing.
    image : `menpo.image.Image` or `subclass` or ``None``, optional
        The image on which the fitting process was applied. Note that a copy
        of the image will be assigned as an attribute. If ``None``, then no
        image is assigned.
    gt_mesh : `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh` or ``None``, optional
        The ground truth mesh associated with the image. If ``None``, then no
        ground truth mesh is assigned.
    model_landmarks_index : `list` or ``None``, optional
        It `list`, then it is supposed to provide indices for vertices of the
        model that have some kind of semantic meaning. These points will be
        used in order to generate 2D pointclouds projected in the image plane.
        If ``None``, then the 2D pointclouds will not be generated.
    """
    def __init__(self, results, affine_transforms, n_scales, image=None,
                 gt_mesh=None, model_landmarks_index=None):
        # Make sure results are iterable with the correct length
        if not isinstance(results, Iterable):
            results = [results]
        if len(results) != n_scales:
            raise ValueError("The provided results and n_scales do not match.")
        # Make sure affine_transforms are iterable with the correct length
        if not isinstance(affine_transforms, Iterable):
            affine_transforms = [affine_transforms]
        if len(affine_transforms) != n_scales:
            raise ValueError("The provided affine_transforms and n_scales do "
                             "not match.")
        # Get initial mesh and initial camera transform
        initial_mesh = None
        if results[0].initial_mesh is not None:
            initial_mesh = results[0].initial_mesh
        initial_camera_transform = None
        if results[0].initial_camera_transform is not None:
            initial_camera_transform = results[0].initial_camera_transform
        # Create meshes list and n_iters_per_scale
        # If the result object has an initial shape, then it has to be
        # removed from the final meshes list
        n_iters_per_scale = []
        meshes = []
        camera_transforms = []
        self._affine_transforms = []
        for i in list(range(n_scales)):
            n_iters_per_scale.append(results[i].n_iters)
            if results[i].initial_mesh is None:
                meshes += results[i].meshes
            else:
                meshes += results[i].meshes[1:]
            if results[i].initial_camera_transform is None:
                camera_transforms += results[i].camera_transforms
                self._affine_transforms += (
                    [_affine_2d_to_3d(affine_transforms[i])] *
                    len(results[i].camera_transforms))
            else:
                camera_transforms += results[i].camera_transforms[1:]
                self._affine_transforms += (
                    [_affine_2d_to_3d(affine_transforms[i])] *
                    len(results[i].camera_transforms[1:]))
        # Call superclass
        super(MultiScaleNonParametricIterativeResult, self).__init__(
            meshes=meshes, camera_transforms=camera_transforms, image=image,
            initial_mesh=initial_mesh,
            initial_camera_transform=initial_camera_transform, gt_mesh=gt_mesh)
        # Get attributes
        self._n_iters_per_scale = n_iters_per_scale
        self._n_scales = n_scales
        # Correct affine transforms, if neccessary
        if self.initial_camera_transform is not None:
            self._affine_transforms = ([self._affine_transforms[0]] +
                                       self._affine_transforms)
        # Create costs list. We assume that if the costs of the first result
        # object is None, then the costs property of all objects is None.
        # Similarly, if the costs property of the the first object is not
        # None, then the same stands for the rest.
        self._costs = None
        if results[0].costs is not None:
            self._costs = []
            for r in results:
                self._costs += r.costs
        # Assing model_landmarks_index
        self._model_landmarks_index = model_landmarks_index

    @property
    def n_iters_per_scale(self):
        r"""
        Returns the number of iterations per scale of the fitting process.

        :type: `list` of `int`
        """
        return self._n_iters_per_scale

    @property
    def n_scales(self):
        r"""
        Returns the number of scales used during the fitting process.

        :type: `int`
        """
        return self._n_scales

    def rasterized_final_mesh(self, shape=None):
        r"""
        Returns the rasterized final mesh. The image's shape will be used in
        case `shape` is not provided.

        Parameters
        ----------
        shape: `(int, int)` or ``None``, optional
            The shape of the rasterized image. If ``None``, then the fitted
            image's shape will be used.

        Returns
        -------
        rasterized_final_mesh : `menpo.image.Image`
            The image with the rasterized final mesh.

        Raises
        ------
        ValueError
            The final camera transform does not exist.
        ValueError
            You need to provide an image shape, since the image does not exist.
        """
        if self.final_camera_transform is None:
            raise ValueError("The final camera transform does not exist.")
        if shape is None:
            if self.image is None:
                raise ValueError("You need to provide an image shape, "
                                 "since the image does not exist.")
            else:
                shape = self.image.shape
        instance_in_img = self._affine_transforms[-1].apply(
            self.final_camera_transform.apply(self.final_mesh))
        return rasterize_mesh(instance_in_img, shape)

    def rasterized_initial_mesh(self, shape=None):
        r"""
        Returns the rasterized initial mesh, if it exists. The image's shape
        will be used in case `shape` is not provided.

        Parameters
        ----------
        shape: `(int, int)` or ``None``, optional
            The shape of the rasterized image. If ``None``, then the fitted
            image's shape will be used.

        Returns
        -------
        rasterized_initial_mesh : `menpo.image.Image`
            The image with the rasterized initial mesh.

        Raises
        ------
        ValueError
            The initial mesh does not exist.
        ValueError
            The initial camera transform does not exist.
        ValueError
            You need to provide an image shape, since the image does not exist.
        """
        if self.initial_mesh is None:
            raise ValueError("The initial mesh does not exist.")
        if self.initial_camera_transform is None:
            raise ValueError("The initial camera transform does not exist.")
        if shape is None:
            if self.image is None:
                raise ValueError("You need to provide an image shape, "
                                 "since the image does not exist.")
            else:
                shape = self.image.shape
        instance_in_img = self._affine_transforms[0].apply(
            self.initial_camera_transform.apply(self.initial_mesh))
        return rasterize_mesh(instance_in_img, shape)

    def rasterized_meshes(self, shape=None):
        r"""
        Returns the rasterized meshes. The image's shape will be used in
        case `shape` is not provided.

        Parameters
        ----------
        shape: `(int, int)` or ``None``, optional
            The shape of the rasterized image. If ``None``, then the fitted
            image's shape will be used.

        Returns
        -------
        rasterized_meshes : `list` of `menpo.image.Image`
            The list of images with the rasterized meshes.

        Raises
        ------
        ValueError
            You need to provide an image shape, since the image does not exist.
        """
        if shape is None:
            if self.image is None:
                raise ValueError("You need to provide an image shape, "
                                 "since the image does not exist.")
            else:
                shape = self.image.shape

        def rast(x):
            affine, camera, mesh = x
            instance_in_img = affine.compose_after(
                camera.camera_transform).apply(mesh)
            return rasterize_mesh(instance_in_img, shape)

        xs = list(zip(self._affine_transforms, self.camera_transforms,
                      self.meshes))
        return LazyList.init_from_iterable(xs, f=rast)

    def sparse_final_mesh_projected_in_2d(self):
        r"""
        Returns the sparse final mesh projected in 2D.

        :type: `menpo.shape.PointCloud`
            The sparse final mesh projected in the image plane.

        Raises
        ------
        ValueError
            There is not a sparse landmarks index mapping.
        ValueError
            The final camera transform does not exist.
        """
        if self._model_landmarks_index is None:
            raise ValueError("There is not a sparse landmarks index mapping.")
        if self.final_camera_transform is None:
            raise ValueError("The final camera transform does not exist.")
        sparse_instance = PointCloud(
            self.final_mesh.points[self._model_landmarks_index])
        instance_in_img = self._affine_transforms[-1].apply(
            self.final_camera_transform.apply(sparse_instance))
        return PointCloud(instance_in_img.points[:, :2])

    def sparse_initial_mesh_projected_in_2d(self):
        r"""
        Returns the sparse initial mesh projected in 2D.

        :type: `menpo.shape.PointCloud`
            The sparse initial mesh projected in the image plane.

        Raises
        ------
        ValueError
            There is not a sparse landmarks index mapping.
        ValueError
            The initial camera transform does not exist.
        """
        if self._model_landmarks_index is None:
            raise ValueError("There is not a sparse landmarks index mapping.")
        if self.initial_mesh is None:
            raise ValueError("The initial mesh does not exist.")
        if self.initial_camera_transform is None:
            raise ValueError("The initial camera transform does not exist.")
        sparse_instance = PointCloud(
            self.initial_mesh.points[self._model_landmarks_index])
        instance_in_img = self._affine_transforms[0].apply(
            self.initial_camera_transform.apply(sparse_instance))
        return PointCloud(instance_in_img.points[:, :2])

    def sparse_meshes_projected_in_2d(self):
        r"""
        Returns the list of sparse meshes projected in 2D.

        :type: `list` of `menpo.shape.PointCloud`
            The list of sparse meshes projected in the image plane.
        """
        def project(x):
            mesh, affine, camera = x
            sparse_instance = PointCloud(
                mesh.points[self._model_landmarks_index])
            instance_in_img = affine.apply(camera.apply(sparse_instance))
            return PointCloud(instance_in_img.points[:, :2])

        xs = list(zip(self.meshes, self._affine_transforms,
                      self.camera_transforms))
        return LazyList.init_from_iterable(xs, f=project)

    def final_mesh_projected_in_2d(self):
        r"""
        Returns the final mesh projected in the image plane.

        :type: `menpo.shape.TriMesh`
            The final mesh projected in the image plane.

        Raises
        ------
        ValueError
            The final camera transform does not exist.
        """
        if self.final_camera_transform is None:
            raise ValueError("The final camera transform does not exist.")
        mesh_in_img = self._affine_transforms[-1].apply(
            self.final_camera_transform.apply(self.final_mesh))
        return as_trimesh(mesh_in_img.with_dims([0, 1]))

    def initial_mesh_projected_in_2d(self):
        r"""
        Returns the initial mesh projected in the image plane.

        :type: `menpo.shape.TriMesh`
            The initial mesh projected in the image plane.

        Raises
        ------
        ValueError
            The initial camera transform does not exist.
        """
        if self.initial_camera_transform is None:
            raise ValueError("The initial camera transform does not exist.")
        mesh_in_img = self._affine_transforms[0].apply(
            self.initial_camera_transform.apply(self.initial_mesh))
        return as_trimesh(mesh_in_img.with_dims([0, 1]))

    def meshes_projected_in_2d(self):
        r"""
        Returns the list of meshes projected in the image plane.

        :type: `list` of `menpo.shape.TriMesh`
            The list of meshes projected in the image plane.
        """
        def project(x):
            mesh, affine, camera = x
            mesh_in_img = affine.apply(camera.apply(mesh))
            return as_trimesh(mesh_in_img.with_dims([0, 1]))

        xs = list(zip(self.meshes, self._affine_transforms,
                      self.camera_transforms))
        return LazyList.init_from_iterable(xs, f=project)

    def final_mesh_with_image_texture(self):
        r"""
        Returns the final mesh with texture that is sampled from the image.

        :type: `menpo.shape.ColouredTriMesh`
            The final mesh with texture sampled from the image.

        Raises
        ------
        ValueError
            The final camera transform does not exist.
        ValueError
            The image does not exist.
        """
        if self.final_camera_transform is None:
            raise ValueError("The final camera transform does not exist.")
        if self.image is None:
            raise ValueError("The image does not exist.")
        mesh_in_img = self._affine_transforms[-1].apply(
            self.final_camera_transform.apply(self.final_mesh))
        colours = extract_per_vertex_colour(mesh_in_img, self.image)
        return as_colouredtrimesh(self.final_mesh, colours=colours)

    def initial_mesh_with_image_texture(self):
        r"""
        Returns the initial mesh with texture that is sampled from the image.

        :type: `menpo.shape.ColouredTriMesh`
            The initial mesh with texture sampled from the image.

        Raises
        ------
        ValueError
            The initial mesh does not exist.
        ValueError
            The initial camera transform does not exist.
        ValueError
            The image does not exist.
        """
        if self.initial_mesh is None:
            raise ValueError("The initial mesh does not exist.")
        if self.initial_camera_transform is None:
            raise ValueError("The initial camera transform does not exist.")
        if self.image is None:
            raise ValueError("The image does not exist.")
        mesh_in_img = self._affine_transforms[0].apply(
            self.initial_camera_transform.apply(self.initial_mesh))
        colours = extract_per_vertex_colour(mesh_in_img, self.image)
        return as_colouredtrimesh(self.initial_mesh, colours=colours)

    def meshes_with_image_texture(self):
        r"""
        Returns the list of meshes per iteration with texture that is sampled
        from the image.

        :type: `list` of `menpo.shape.ColouredTriMesh`
            The list of meshes with texture sampled from the image.

        Raises
        ------
        ValueError
            The image does not exist.
        """
        if self.image is None:
            raise ValueError("The image does not exist.")

        def mesh_with_image_texture(x):
            mesh, affine, camera = x
            mesh_in_img = affine.apply(camera.apply(mesh))
            colours = extract_per_vertex_colour(mesh_in_img, self.image)
            return as_colouredtrimesh(mesh, colours=colours)

        xs = list(zip(self.meshes, self._affine_transforms,
                      self.camera_transforms))
        return LazyList.init_from_iterable(xs, f=mesh_with_image_texture)


class MultiScaleParametricIterativeResult(MultiScaleNonParametricIterativeResult):
    r"""
    Class for defining a multi-scale parametric iterative fitting result, i.e.
    the result of a multi-scale method that optimizes over a parametric shape
    model. It holds the meshes of all the iterations of the fitting procedure,
    as well as the scales. It can optionally store the image on which the
    fitting was applied, as well as its ground truth mesh.

    .. note:: When using a method with a parametric shape model, the first step
              is to **reconstruct the initial mesh** using the shape model. The
              generated reconstructed mesh is then used as initialisation for
              the iterative optimisation. This step is not counted in the number
              of iterations.

    Parameters
    ----------
    results : `list` of :map:`ParametricIterativeResult`
        The `list` of non parametric iterative results per scale.
    n_scales : `int`
        The number of scales.
    affine_transforms : `list` of `menpo.transform.Affine`
        The list of affine transforms per scale that transform where used
        dutring the image's pre-processing.
    image : `menpo.image.Image` or `subclass` or ``None``, optional
        The image on which the fitting process was applied. Note that a copy
        of the image will be assigned as an attribute. If ``None``, then no
        image is assigned.
    gt_mesh : `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh` or ``None``, optional
        The ground truth mesh associated with the image. If ``None``, then no
        ground truth mesh is assigned.
    model_landmarks_index : `list` or ``None``, optional
        It `list`, then it is supposed to provide indices for vertices of the
        model that have some kind of semantic meaning. These points will be
        used in order to generate 2D pointclouds projected in the image plane.
        If ``None``, then the 2D pointclouds will not be generated.
    """
    def __init__(self, results, affine_transforms, n_scales, image=None,
                 gt_mesh=None, model_landmarks_index=None):
        # Call superclass
        super(MultiScaleParametricIterativeResult, self).__init__(
            results=results, affine_transforms=affine_transforms,
            n_scales=n_scales, image=image, gt_mesh=gt_mesh,
            model_landmarks_index=model_landmarks_index)
        # Create shape parameters
        self._shape_parameters = []
        for r in results:
            self._shape_parameters += r.shape_parameters
        # Correct n_iters
        self._n_iters -= n_scales

    @property
    def shape_parameters(self):
        r"""
        Returns the `list` of shape parameters obtained at each iteration of
        the fitting process. The `list` includes the parameters of the
        `initial_mesh` (if it exists) and `final_mesh`.

        :type: `list` of ``(n_params,)`` `ndarray`
        """
        return self._shape_parameters

    @property
    def reconstructed_initial_meshes(self):
        r"""
        Returns the result of the reconstruction step that takes place at each
        scale before applying the iterative optimisation.

        :type: `list` of `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh`
        """
        ids = self._reconstruction_indices
        return [self.meshes[i] for i in ids]

    @property
    def reconstructed_initial_camera_transforms(self):
        r"""
        Returns the camera transforms of the reconstructed initial meshes.

        :type: `list` of `menpo3d.camera.PerspectiveCamera`
        """
        ids = self._reconstruction_indices
        return [self.camera_transforms[i] for i in ids]

    def rasterized_reconstructed_initial_meshes(self, shape=None):
        r"""
        Returns the rasterized reconstructed initial meshes. The image's shape
        will be used in case `shape` is not provided.

        Parameters
        ----------
        shape: `(int, int)` or ``None``, optional
            The shape of the rasterized image. If ``None``, then the fitted
            image's shape will be used.

        Returns
        -------
        rasterized_reconstructed_initial_meshes : `list` of `menpo.image.Image`
            The list of images with the rasterized reconstructed initial meshes.

        Raises
        ------
        ValueError
            You need to provide an image shape, since the image does not exist.
        """
        if shape is None:
            if self.image is None:
                raise ValueError("You need to provide an image shape, "
                                 "since the image does not exist.")
            else:
                shape = self.image.shape
        ids = self._reconstruction_indices
        rasterized_meshes = []
        for i in ids:
            instance_in_img = self._affine_transforms[i].apply(
                self.camera_transforms[i].apply(self.meshes[i]))
            rasterized_meshes.append(rasterize_mesh(instance_in_img, shape))
        return rasterized_meshes

    @property
    def _reconstruction_indices(self):
        r"""
        Returns a list with the indices of reconstructed meshes in the `meshes`
        list.

        :type: `list` of `int`
        """
        initial_val = 0
        if self.initial_mesh is not None:
            initial_val = 1
        ids = []
        for i in list(range(self.n_scales)):
            if i == 0:
                ids.append(initial_val)
            else:
                previous_val = ids[i - 1]
                ids.append(previous_val + self.n_iters_per_scale[i - 1] + 1)
        return ids

    def reconstructed_initial_error(self, compute_error=None):
        r"""
        Returns the error of the reconstructed initial mesh of the fitting
        process, if the ground truth mesh exists. This is the error computed
        based on the `reconstructed_initial_meshes[0]`.

        Parameters
        ----------
        compute_error: `callable`, optional
            Callable that computes the error between the reconstructed initial
            and ground truth meshes.

        Returns
        -------
        reconstructed_initial_error : `float`
            The error that corresponds to the initial mesh's reconstruction.

        Raises
        ------
        ValueError
            Ground truth mesh has not been set, so the reconstructed initial
            error cannot be computed
        """
        if compute_error is None:
            compute_error = error_function
        if self.gt_mesh is None:
            raise ValueError('Ground truth mesh has not been set, so the '
                             'reconstructed initial error cannot be computed')
        else:
            return compute_error(self.reconstructed_initial_meshes[0],
                                 self.gt_mesh)



from menpo.io.output.extensions import ljson_exporter
from .mesh import obj_exporter, ply_exporter


landmark_types = {
    '.ljson': ljson_exporter
}

mesh_types_buffer_support = {
    '.obj': obj_exporter
}

mesh_types_paths_only = {
    '.ply': ply_exporter
}

mesh_types = {}
mesh_types.update(mesh_types_buffer_support)
mesh_types.update(mesh_types_paths_only)


from menpo3d.result import (ParametricIterativeResult,
                            MultiScaleParametricIterativeResult)


class MMAlgorithmResult(ParametricIterativeResult):
    r"""
    Class for storing the iterative result of a MM optimisation algorithm.

    .. note:: When using a method with a parametric shape model, the first step
              is to **reconstruct the initial mesh** using the shape model. The
              generated reconstructed mesh is then used as initialisation for
              the iterative optimisation. This step is not counted in the number
              of iterations.

    Parameters
    ----------
    shape_parameters : `list` of `ndarray`
        The `list` of shape parameters per iteration. Note that the list
        includes the parameters of the projection of the initial mesh. The last
        member of the list corresponds to the final mesh's parameters. It must
        have the same length as `camera_parameters` and `meshes`.
    texture_parameters : `list` of `ndarray`
        The `list` of texture parameters per iteration. Note that the list
        includes the parameters of the projection of the initial mesh. The last
        member of the list corresponds to the final mesh's parameters. It must
        have the same length as `camera_transforms` and `meshes`.
    meshes : `list` of `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh`
        The `list` of meshes per iteration. Note that the list does not
        include the initial mesh. However, it includes the reconstruction of
        the initial mesh. The last member of the list is the final mesh.
    camera_transforms : `list` of `menpo3d.camera.PerspectiveCamera`
        The `list` of camera transform objects per iteration. Note that the
        list does not include the initial camera transform. The last member
        of the list is the final camera transform.
    image : `menpo.image.Image` or `subclass` or ``None``, optional
        The image on which the fitting process was applied. Note that a copy
        of the image will be assigned as an attribute. If ``None``, then no
        image is assigned.
    initial_mesh : `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh` or ``None``, optional
        The initial mesh from which the fitting process started. If
        ``None``, then no initial mesh is assigned.
    initial_camera_transform : `menpo3d.camera.PerspectiveCamera` or ``None``, optional
        The initial camera transform. If ``None``, then no initial camera
        is assigned.
    gt_mesh : `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh` or ``None``, optional
        The ground truth mesh associated with the image. If ``None``, then
        no ground truth mesh is assigned.
    costs : `list` of `float` or ``None``, optional
        The `list` of cost per iteration. If ``None``, then it is assumed
        that the cost function cannot be computed for the specific
        algorithm. It must have the same length as `meshes`.
    """
    def __init__(self, shape_parameters, texture_parameters, meshes,
                 camera_transforms, image=None, initial_mesh=None,
                 initial_camera_transform=None, gt_mesh=None, costs=None):
        super(MMAlgorithmResult, self).__init__(
            shape_parameters=shape_parameters, meshes=meshes,
            camera_transforms=camera_transforms, image=image,
            initial_mesh=initial_mesh,
            initial_camera_transform=initial_camera_transform, gt_mesh=gt_mesh,
            costs=costs)
        self._texture_parameters = texture_parameters

    @property
    def texture_parameters(self):
        r"""
        Returns the `list` of texture parameters obtained at each iteration
        of the fitting process. The `list` includes the parameters of the
        `initial_mesh` (if it exists) and `final_mesh`.

        :type: `list` of ``(n_params,)`` `ndarray`
        """
        return self._texture_parameters


class MMResult(MultiScaleParametricIterativeResult):
    r"""
    Class for storing the multi-scale iterative fitting result of a MM. It
    holds the meshes, shape parameters, texture parameters and costs per
    iteration.

    .. note:: When using a method with a parametric shape model, the first step
              is to **reconstruct the initial mesh** using the shape model. The
              generated reconstructed mesh is then used as initialisation for
              the iterative optimisation. This step is not counted in the number
              of iterations.

    Parameters
    ----------
    results : `list` of :map:`ParametricIterativeResult`
        The `list` of non parametric iterative results per scale.
    n_scales : `int`
        The number of scales.
    affine_transforms : `list` of `menpo.transform.Affine`
        The list of affine transforms per scale that transform where used
        dutring the image's pre-processing.
    image : `menpo.image.Image` or `subclass` or ``None``, optional
        The image on which the fitting process was applied. Note that a copy
        of the image will be assigned as an attribute. If ``None``, then no
        image is assigned.
    gt_mesh : `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh` or ``None``, optional
        The ground truth mesh associated with the image. If ``None``, then no
        ground truth mesh is assigned.
    model_landmarks_index : `list` or ``None``, optional
        It `list`, then it is supposed to provide indices for vertices of the
        model that have some kind of semantic meaning. These points will be
        used in order to generate 2D pointclouds projected in the image plane.
        If ``None``, then the 2D pointclouds will not be generated.
    """
    def __init__(self, results, affine_transforms, n_scales, image=None,
                 gt_mesh=None, model_landmarks_index=None):
        super(MMResult, self).__init__(
            results=results, affine_transforms=affine_transforms,
            n_scales=n_scales, image=image, gt_mesh=gt_mesh,
            model_landmarks_index=model_landmarks_index)
        # Create texture parameters list
        self._texture_parameters = None
        if results[0].texture_parameters is not None:
            self._texture_parameters = []
            for r in results:
                self._texture_parameters += r.texture_parameters

    @property
    def texture_parameters(self):
        r"""
        Returns the `list` of texture parameters obtained at each iteration
        of the fitting process. The `list` includes the parameters of the
        `initial_mesh` (if it exists) and `final_mesh`.

        :type: `list` of ``(n_params,)`` `ndarray`
        """
        return self._texture_parameters



from functools import reduce
import numpy as np
from menpo.transform import Homogeneous, Translation, Scale, NonUniformScale


def dims_3to2():
    r"""

    Returns
     ------
    :map`Homogeneous`
        :map`Homogeneous` that strips off the 3D axis of a 3D shape
        leaving just the first two axes.

    """
    return Homogeneous(np.array([[1, 0, 0, 0],
                                 [0, 1, 0, 0],
                                 [0, 0, 0, 1]]))


def dims_2to3(x=0):
    r"""
    Return a :map`Homogeneous` that adds on a 3rd axis to a 2D shape.

    Parameters
    ----------
    x : `float`, optional
        The value that will be assigned to the new third dimension

    Returns
     ------
    :map`Homogeneous`
        :map`Homogeneous` that adds on a 3rd axis to a 2D shape.

    """
    return Homogeneous(np.array([[1, 0, 0],
                                 [0, 1, 0],
                                 [0, 0, x],
                                 [0, 0, 1]]))


def model_to_clip_transform(points, xy_scale=0.9, z_scale=0.3):
    r"""
    Produces an Affine Transform which centres and scales 3D points to fit
    into the OpenGL clipping space ([-1, 1], [-1, 1], [1, 1-]). This can be
    used to construct an appropriate projection matrix for use in an
    orthographic Rasterizer. Note that the z-axis is flipped as is default in
    OpenGL - as a result this transform converts the right handed coordinate
    input into a left hand one.

    Parameters
    ----------

    points: :map:`PointCloud`
        The points that should be adjusted.

    xy_scale: `float` 0-1, optional
        Amount by which the boundary is relaxed so the points are not
        right against the edge. A value of 1 means the extremities of the
        point cloud will be mapped onto [-1, 1] [-1, 1] exactly (no boarder)
        A value of 0.5 means the points will be mapped into the range
        [-0.5, 0.5].

        Default: 0.9 (map to [-0.9, 0.9])

    z_scale: float 0-1, optional
        Scale factor by which the z-dimension is squeezed. A value of 1
        means the z-range of the points will be mapped to exactly fit in
        [1, -1]. A scale of 0.1 means the z-range is compressed to fit in the
        range [0.1, -0.1].

    Returns
    -------
    :map:`Affine`
        The affine transform that creates this mapping
    """
    # 1. Centre the points on the origin
    center = Translation(points.centre_of_bounds()).pseudoinverse()
    # 2. Scale the points to exactly fit the boundaries
    scale = Scale(points.range() / 2.0)
    # 3. Apply the relaxations requested - note the flip in the z axis!!
    # This is because OpenGL by default evaluates depth as bigger number ==
    # further away. Thus not only do we need to get to clip space [-1, 1] in
    # all dims) but we must invert the z axis so depth buffering is correctly
    # applied.
    b_scale = NonUniformScale([xy_scale, xy_scale, -z_scale])
    return center.compose_before(scale.pseudoinverse()).compose_before(b_scale)


def clip_to_image_transform(width, height):
    r"""
    Affine transform that converts 3D clip space coordinates into 2D image
    space coordinates. Note that the z axis of the clip space coordinates is
    ignored.

    Parameters
    ----------

    width: int
        The width of the image

    height: int
        The height of the image

    Returns
    -------

    HomogeneousTransform
        A homogeneous transform that moves clip space coordinates into image
        space.
    """
    # 1. Remove the z axis from the clip space
    rem_z = dims_3to2()
    # 2. invert the y direction (up becomes down)
    invert_y = Scale([1, -1])
    # 3. [-1, 1] [-1, 1] -> [0, 2] [0, 2]
    t = Translation([1, 1])
    # 4. [0, 2] [0, 2] -> [0, 1] [0, 1]
    unit_scale = Scale(0.5, n_dims=2)
    # 5. [0, 1] [0, 1] -> [0, w - 1] [0, h - 1]
    im_scale = Scale([width - 1, height - 1])
    # 6. [0, w] [0, h] -> [0, h] [0, w]
    xy_yx = Homogeneous(np.array([[0, 1, 0],
                                  [1, 0, 0],
                                  [0, 0, 1]], dtype=np.float))
    # reduce the full transform chain to a single affine matrix
    transforms = [rem_z, invert_y, t, unit_scale, im_scale, xy_yx]
    return reduce(lambda a, b: a.compose_before(b), transforms)



import numpy as np
import scipy.sparse as sp
from menpo.shape import TriMesh, PointCloud
from menpo.transform import Translation, UniformScale, AlignmentSimilarity
from menpo3d.vtkutils import trimesh_to_vtk, VTKClosestPointLocator
from menpo3d.morphablemodel.shapemodel import ShapeModel


import os
import sys
from contextlib import contextmanager
import warnings

@contextmanager
def stdout_redirected(to=os.devnull):
    r"""
    import os

    with stdout_redirected(to=filename):
        print("from Python")
        os.system("echo non-Python applications are also supported")
    """
    fd = sys.stdout.fileno()

    # assert that Python and C stdio write using the same file descriptor
    # assert libc.fileno(ctypes.c_void_p.in_dll(libc, "stdout")) == fd == 1

    def redirect_stdout(to):
        sys.stdout.close()  # + implicit flush()
        os.dup2(to.fileno(), fd)  # fd writes to 'to' file
        sys.stdout = os.fdopen(fd, 'w')  # Python writes to fd

    with os.fdopen(os.dup(fd), 'w') as old_stdout:
        with open(to, 'w') as file:
            redirect_stdout(to=file)
        try:
            yield  # allow code to be run with the redirected stdout
        finally:
            # restore stdout.
            # buffering and flags such as CLOEXEC may be different
            redirect_stdout(to=old_stdout)

try:
    
    try:
        # First try the newer scikit-sparse namespace
        from sksparse.cholmod import cholesky_AAt
    except ImportError:
        # Fall back to the older scikits.sparse namespace
        from scikits.sparse.cholmod import cholesky_AAt

    # user has cholesky available - provide a fast solve
    def spsolve(sparse_X, dense_b):
        # wrap the cholesky call in a context manager that swallows the
        # low-level std-out to stop it from swamping our stdout (these low-level
        # prints come from METIS, but the solution behaves as normal)
        with stdout_redirected():
            factor = cholesky_AAt(sparse_X.T)
        return factor(sparse_X.T.dot(dense_b)).toarray()

except ImportError:
    # fallback to (much slower) scipy solve
    from scipy.sparse.linalg import spsolve as scipy_spsolve

    def spsolve(sparse_X, dense_b):
        warnings.warn("suitesparse is not installed - NICP will run "
                      "considerably (~5-10x) slower. If possible install "
                      "suitesparse.")
        return scipy_spsolve(sparse_X.T.dot(sparse_X),
                             sparse_X.T.dot(dense_b)).toarray()


def node_arc_incidence_matrix(source):
    unique_edge_pairs = source.unique_edge_indices()
    m = unique_edge_pairs.shape[0]

    # Generate a "node-arc" (i.e. vertex-edge) incidence matrix.
    row = np.hstack((np.arange(m), np.arange(m)))
    col = unique_edge_pairs.T.ravel()
    data = np.hstack((-1 * np.ones(m), np.ones(m)))
    return sp.coo_matrix((data, (row, col))), unique_edge_pairs


def validate_weights(label, weights, n_points, n_iterations=None,
                     verbose=False):
    if n_iterations is not None and len(weights) != n_iterations:
        raise ValueError('Invalid {label}: - due to other weights there are '
                         '{n_iterations} iterations but {n_weights} {label} '
                         'were provided'.format(label=label,
                                                n_iterations=n_iterations,
                                                n_weights=len(weights)))
    invalid = []
    for i, weight in enumerate(weights):
        is_per_vertex = isinstance(weight, np.ndarray)
        if is_per_vertex and weight.shape != (n_points,):
                invalid.append('({}): {}'.format(i, weight.shape[0]))

    if verbose and len(weights) >= 1:
        is_per_vertex = isinstance(weights[0], np.ndarray)
        if is_per_vertex:
            print('Using per-vertex {label}'.format(label=label))
        else:
            print('Using global {label}'.format(label=label))

    if len(invalid) != 0:
        raise ValueError('Invalid {label}: expected shape ({n_points},) '
                         'got: {invalid_cases}'.format(
                            label=label, n_points=n_points,
                            invalid_cases='{}'.format(', '.join(invalid))))


def non_rigid_icp(source, target, eps=1e-3, landmark_group=None,
                  stiffness_weights=None, data_weights=None,
                  landmark_weights=None, generate_instances=False,
                  verbose=False):
    # call the generator version of NICP, always returning a generator
    generator = non_rigid_icp_generator(source, target, eps=eps,
                                        stiffness_weights=stiffness_weights,
                                        verbose=verbose,
                                        landmark_group=landmark_group,
                                        landmark_weights=landmark_weights,
                                        data_weights=data_weights)
    # the handler decides whether the user get's details and each iteration
    # returned, or just the final result.
    return non_rigid_icp_generator_handler(generator, generate_instances)


def active_non_rigid_icp(model, target, eps=1e-3,
                         stiffness_weights=None, data_weights=None,
                         landmark_group=None, landmark_weights=None,
                         model_mean_landmarks=None,
                         generate_instances=False, verbose=False):

    model_mean = model.mean()

    if landmark_group is not None:

        # user better have provided model landmarks!
        if model_mean_landmarks is None:
            raise ValueError(
                'For Active NICP with landmarks the model_mean_landmarks '
                'need to be provided.')

        shape_model = ShapeModel(model)
        source_lms = model_mean_landmarks
        target_lms = target.landmarks[landmark_group]
        model_lms_index = model_mean.distance_to(source_lms).argmin(axis=0)
        shape_model_lms = shape_model.mask_points(model_lms_index)

        # Sim align the target lms to the mean before projecting
        target_lms_aligned = AlignmentSimilarity(target_lms,
                                                 source_lms).apply(target_lms)

        # project to learn the weights for the landmark model
        weights = shape_model_lms.project(target_lms_aligned,
                                          n_components=20)

        # use these weights on the dense shape model to generate an improved
        # instance
        source = model.instance(weights)

        # update the source landmarks (for the alignment below)
        source.landmarks[landmark_group] = PointCloud(source.points[
                                                  model_lms_index])
    else:
        # Start from the mean of the model
        source = model_mean

    # project onto the shape model to restrict the basis
    def project_onto_model(instance):
        return model.reconstruct(instance)
    # call the generator version of NICP, always returning a generator
    generator = non_rigid_icp_generator(source, target, eps=eps,
                                        stiffness_weights=stiffness_weights,
                                        data_weights=data_weights,
                                        landmark_weights=landmark_weights,
                                        landmark_group=landmark_group,
                                        v_i_update_func=project_onto_model,
                                        verbose=verbose)
    # the handler decides whether the user get's details and each iteration
    # returned, or just the final result.
    return non_rigid_icp_generator_handler(generator, generate_instances)


def non_rigid_icp_generator_handler(generator, generate_instances):

    if generate_instances:
        # the user wants to inspect results per-iteration - return the iterator
        # directly to them
        return generator
    else:
        # the user is not interested in per-iteration results. Exhaust the
        # generator ourselves and return the last result only.
        while True:
            try:
                instance = next(generator)
            except StopIteration:
                return instance[0]


def non_rigid_icp_generator(source, target, eps=1e-3,
                            stiffness_weights=None, data_weights=None,
                            landmark_group=None, landmark_weights=None,
                            v_i_update_func=None, verbose=False):
    r"""
    Deforms the source trimesh to align with to optimally the target.
    """
    # If landmarks are provided, we should always start with a simple
    # AlignmentSimilarity between the landmarks to initialize optimally.
    if landmark_group is not None:
        if verbose:
            print("'{}' landmarks will be used as "
                  "a landmark constraint.".format(landmark_group))
            print("performing similarity alignment using landmarks")
        lm_align = AlignmentSimilarity(source.landmarks[landmark_group],
                                       target.landmarks[landmark_group]).as_non_alignment()
        source = lm_align.apply(source)

    # Scale factors completely change the behavior of the algorithm - always
    # rescale the source down to a sensible size (so it fits inside box of
    # diagonal 1) and is centred on the origin. We'll undo this after the fit
    # so the user can use whatever scale they prefer.
    tr = Translation(-1 * source.centre())
    sc = UniformScale(1.0 / np.sqrt(np.sum(source.range() ** 2)), 3)
    prepare = tr.compose_before(sc)

    source = prepare.apply(source)
    target = prepare.apply(target)

    # store how to undo the similarity transform
    restore = prepare.pseudoinverse()

    n_dims = source.n_dims
    # Homogeneous dimension (1 extra for translation effects)
    h_dims = n_dims + 1
    points, trilist = source.points, source.trilist
    n = points.shape[0]  # record number of points

    edge_tris = source.boundary_tri_index()

    M_s, unique_edge_pairs = node_arc_incidence_matrix(source)

    # weight matrix
    G = np.identity(n_dims + 1)

    M_kron_G_s = sp.kron(M_s, G)

    # build octree for finding closest points on target.
    target_vtk = trimesh_to_vtk(target)
    closest_points_on_target = VTKClosestPointLocator(target_vtk)

    # save out the target normals. We need them for the weight matrix.
    target_tri_normals = target.tri_normals()

    # init transformation
    X_prev = np.tile(np.zeros((n_dims, h_dims)), n).T
    v_i = points

    if stiffness_weights is not None:
        if verbose:
            print('using user-defined stiffness_weights')
        validate_weights('stiffness_weights', stiffness_weights,
                         source.n_points, verbose=verbose)
    else:
        # these values have been empirically found to perform well for well
        # rigidly aligned facial meshes
        stiffness_weights = [50, 20, 5, 2, 0.8, 0.5, 0.35, 0.2]
        if verbose:
            print('using default '
                  'stiffness_weights: {}'.format(stiffness_weights))

    n_iterations = len(stiffness_weights)

    if landmark_weights is not None:
        if verbose:
            print('using user defined '
                  'landmark_weights: {}'.format(landmark_weights))
    elif landmark_group is not None:
        # these values have been empirically found to perform well for well
        # rigidly aligned facial meshes
        landmark_weights = [5, 2, .5, 0, 0, 0, 0, 0]
        if verbose:
            print('using default '
                  'landmark_weights: {}'.format(landmark_weights))
    else:
        # no landmark_weights provided - no landmark_group in use. We still
        # need a landmark group for the iterator
        landmark_weights = [None] * n_iterations

    # We should definitely have some landmark weights set now - check the
    # number is correct.
    # Note we say verbose=False, as we have done custom reporting above, and
    # per-vertex landmarks are not supported.
    validate_weights('landmark_weights', landmark_weights, source.n_points,
                     n_iterations=n_iterations, verbose=False)

    if data_weights is not None:
        if verbose:
            print('using user-defined data_weights')
        validate_weights('data_weights', data_weights,
                         source.n_points, n_iterations=n_iterations,
                         verbose=verbose)
    else:
        data_weights = [None] * n_iterations
        if verbose:
            print('Not customising data_weights')

    # we need to prepare some indices for efficient construction of the D
    # sparse matrix.
    row = np.hstack((np.repeat(np.arange(n)[:, None], n_dims, axis=1).ravel(),
                     np.arange(n)))

    x = np.arange(n * h_dims).reshape((n, h_dims))
    col = np.hstack((x[:, :n_dims].ravel(),
                     x[:, n_dims]))
    o = np.ones(n)

    if landmark_group is not None:
        source_lm_index = source.distance_to(
            source.landmarks[landmark_group]).argmin(axis=0)
        target_lms = target.landmarks[landmark_group]
        U_L = target_lms.points
        n_landmarks = target_lms.n_points
        lm_mask = np.in1d(row, source_lm_index)
        col_lm = col[lm_mask]
        # pull out the rows for the lms - but the values are
        # all wrong! need to map them back to the order of the landmarks
        row_lm_to_fix = row[lm_mask]
        source_lm_index_l = list(source_lm_index)
        row_lm = np.array([source_lm_index_l.index(r) for r in row_lm_to_fix])

    for i, (alpha, beta, gamma) in enumerate(zip(stiffness_weights,
                                                 landmark_weights,
                                                 data_weights), 1):
        alpha_is_per_vertex = isinstance(alpha, np.ndarray)
        if alpha_is_per_vertex:
            # stiffness is provided per-vertex
            if alpha.shape[0] != source.n_points:
                raise ValueError()
            alpha_per_edge = alpha[unique_edge_pairs].mean(axis=1)
            alpha_M_s = sp.diags(alpha_per_edge).dot(M_s)
            alpha_M_kron_G_s = sp.kron(alpha_M_s, G)
        else:
            # stiffness is global - just a scalar multiply. Note that here
            # we don't have to recalculate M_kron_G_s
            alpha_M_kron_G_s = alpha * M_kron_G_s

        if verbose:
            a_str = (alpha if not alpha_is_per_vertex
                     else 'min: {:.2f}, max: {:.2f}'.format(alpha.min(),
                                                            alpha.max()))
            i_str = '{}/{}: stiffness: {}'.format(i, len(stiffness_weights), a_str)
            if landmark_group is not None:
                i_str += '  lm_weight: {}'.format(beta)
            print(i_str)

        j = 0
        while True:  # iterate until convergence
            j += 1  # track the iterations for this alpha/landmark weight

            # find nearest neighbour and the normals
            U, tri_indices = closest_points_on_target(v_i)

            # ---- WEIGHTS ----
            # 1.  Edges
            # Are any of the corresponding tris on the edge of the target?
            # Where they are we return a false weight (we *don't* want to
            # include these points in the solve)
            w_i_e = np.in1d(tri_indices, edge_tris, invert=True)

            # 2. Normals
            # Calculate the normals of the current v_i
            v_i_tm = TriMesh(v_i, trilist=trilist)
            v_i_n = v_i_tm.vertex_normals()
            # Extract the corresponding normals from the target
            u_i_n = target_tri_normals[tri_indices]
            # If the dot of the normals is lt 0.9 don't contrib to deformation
            w_i_n = (u_i_n * v_i_n).sum(axis=1) > 0.9

            # 3. Self-intersection
            # This adds approximately 12% to the running cost and doesn't seem
            # to be very critical in helping mesh fitting performance so for
            # now it's removed. Revisit later.
            # # Build an intersector for the current deformed target
            # intersect = build_intersector(to_vtk(v_i_tm))
            # # budge the source points 1% closer to the target
            # source = v_i + ((U - v_i) * 0.5)
            # # if the vector from source to target intersects the deformed
            # # template we don't want to include it in the optimisation.
            # problematic = [i for i, (s, t) in enumerate(zip(source, U))
            #                if len(intersect(s, t)[0]) > 0]
            # print(len(problematic) * 1.0 / n)
            # w_i_i = np.ones(v_i_tm.n_points, dtype=np.bool)
            # w_i_i[problematic] = False

            # Form the overall w_i from the normals, edge case
            # for now disable the edge constraint (it was noisy anyway)
            w_i = w_i_n

            # w_i = np.logical_and(w_i_n, w_i_e).astype(np.float)

            # we could add self intersection at a later date too...
            # w_i = np.logical_and(np.logical_and(w_i_n,
            #                                     w_i_e,
            #                                     w_i_i).astype(np.float)

            prop_w_i = (n - w_i.sum() * 1.0) / n
            prop_w_i_n = (n - w_i_n.sum() * 1.0) / n
            prop_w_i_e = (n - w_i_e.sum() * 1.0) / n

            if gamma is not None:
                w_i = w_i * gamma

            # Build the sparse diagonal weight matrix
            W_s = sp.diags(w_i.astype(np.float)[None, :], [0])

            data = np.hstack((v_i.ravel(), o))
            D_s = sp.coo_matrix((data, (row, col)))

            to_stack_A = [alpha_M_kron_G_s, W_s.dot(D_s)]
            to_stack_B = [np.zeros((alpha_M_kron_G_s.shape[0], n_dims)),
                          U * w_i[:, None]]  # nullify nearest points by w_i

            if landmark_group is not None:
                D_L = sp.coo_matrix((data[lm_mask], (row_lm, col_lm)),
                                    shape=(n_landmarks, D_s.shape[1]))
                to_stack_A.append(beta * D_L)
                to_stack_B.append(beta * U_L)

            A_s = sp.vstack(to_stack_A).tocsr()
            B_s = sp.vstack(to_stack_B).tocsr()
            X = spsolve(A_s, B_s)

            # deform template
            v_i_prev = v_i
            v_i = D_s.dot(X)
            delta_v_i = v_i - v_i_prev

            if v_i_update_func:
                # custom logic is provided to update the current template
                # deformation. This is typically used by Active NICP.

                # take the v_i points matrix and convert back to a TriMesh in
                # the original space
                def_template = restore.apply(source.from_vector(v_i.ravel()))

                # perform the update
                updated_def_template = v_i_update_func(def_template)

                # convert back to points in the NICP space
                v_i = prepare.apply(updated_def_template.points)

            err = np.linalg.norm(X_prev - X, ord='fro')
            stop_criterion = err / np.sqrt(np.size(X_prev))

            if landmark_group is not None:
                src_lms = v_i[source_lm_index]
                lm_err = np.sqrt((src_lms - U_L) ** 2).sum(axis=1).mean()

            if verbose:
                v_str = (' - {} stop crit: {:.3f}  '
                         'total: {:.0%}  norms: {:.0%}  '
                         'edges: {:.0%}'.format(j, stop_criterion,
                                                prop_w_i, prop_w_i_n,
                                                prop_w_i_e))
                if landmark_group is not None:
                    v_str += '  lm_err: {:.4f}'.format(lm_err)

                print(v_str)

            X_prev = X

            # track the progress of the algorithm per-iteration
            info_dict = {
                'alpha': alpha,
                'iteration': j,
                'prop_omitted': prop_w_i,
                'prop_omitted_norms': prop_w_i_n,
                'prop_omitted_edges': prop_w_i_e,
                'delta': err,
                'mask_normals': w_i_n,
                'mask_edges': w_i_e,
                'mask_all': w_i,
                'nearest_points': restore.apply(U),
                'deformation_per_step': delta_v_i
            }

            current_instance = source.copy()
            current_instance.points = v_i.copy()

            if landmark_group:
                info_dict['beta'] = beta
                info_dict['lm_err'] = lm_err
                current_instance.landmarks[landmark_group] = PointCloud(src_lms)

            yield restore.apply(current_instance), info_dict

            if stop_criterion < eps:
                break



from collections import OrderedDict
from functools import partial
import numpy as np
from scipy.sparse import csr_matrix

from menpo.io.input.landmark import pts_importer
from menpo.shape import PointCloud, LabelledPointUndirectedGraph


pts_mesh_importer = partial(pts_importer, image_origin=False)
pts_mesh_importer = pts_importer.__doc__


def lm3_importer(filepath, asset=None, **kwargs):
    r"""
    Importer for the LM3 file format from the bosphorus dataset. This is a 3D
    landmark type and so it is assumed it only applies to meshes.

    Landmark set label: LM3

    Landmark labels:

    +------------------------+
    | label                  |
    +========================+
    | outer_left_eyebrow     |
    | middle_left_eyebrow    |
    | inner_left_eyebrow     |
    | inner_right_eyebrow    |
    | middle_right_eyebrow   |
    | outer_right_eyebrow    |
    | outer_left_eye_corner  |
    | inner_left_eye_corner  |
    | inner_right_eye_corner |
    | outer_right_eye_corner |
    | nose_saddle_left       |
    | nose_saddle_right      |
    | left_nose_peak         |
    | nose_tip               |
    | right_nose_peak        |
    | left_mouth_corner      |
    | upper_lip_outer_middle |
    | right_mouth_corner     |
    | upper_lip_inner_middle |
    | lower_lip_inner_middle |
    | lower_lip_outer_middle |
    | chin_middle            |
    +------------------------+

    Parameters
    ----------
    filepath : `Path`
        Absolute filepath of the file.
    asset : `object`, optional
        An optional asset that may help with loading. This is unused for this
        implementation.
    \**kwargs : `dict`, optional
        Any other keyword arguments.

    Returns
    -------
    landmarks : :map:`LabelledPointUndirectedGraph`
        The landmarks including appropriate labels.
    """
    with open(str(filepath), 'r') as f:
        landmarks = f.read()

    # Remove comments and blank lines
    landmark_text = [l for l in landmarks.splitlines()
                     if (l.rstrip() and '#' not in l)]

    # First line says how many landmarks there are: 24 Landmarks
    # So pop it off the front
    num_points = int(landmark_text.pop(0).split()[0])
    xs = []
    ys = []
    zs = []
    labels = []

    # The lines then alternate between the labels and the coordinates
    for i in range(num_points * 2):
        if i % 2 == 0:  # label
            # Lowercase, remove spaces and replace with underscores
            l = landmark_text[i]
            l = '_'.join(l.lower().split())
            labels.append(l)
        else:  # coordinate
            p = landmark_text[i].split()
            xs.append(float(p[0]))
            ys.append(float(p[1]))
            zs.append(float(p[2]))

    xs = np.array(xs, dtype=np.float).reshape((-1, 1))
    ys = np.array(ys, dtype=np.float).reshape((-1, 1))
    zs = np.array(zs, dtype=np.float).reshape((-1, 1))

    points = np.hstack([xs, ys, zs])
    # Create the mask whereby there is one landmark per label
    # (identity matrix)
    masks = np.eye(num_points).astype(np.bool)
    masks = np.vsplit(masks, num_points)
    masks = [np.squeeze(m) for m in masks]
    labels_to_masks = OrderedDict(zip(labels, masks))

    empty_adj_matrix = csr_matrix((num_points, num_points))
    return LabelledPointUndirectedGraph(points, empty_adj_matrix,
                                        labels_to_masks)


def lan_importer(filepath, asset=None, **kwargs):
    r"""
    Importer for the LAN file format for the GOSH dataset. This is a 3D
    landmark type and so it is assumed it only applies to meshes.

    Landmark set label: LAN

    Note that the exact meaning of each landmark in this set varies,
    so all we can do is import all landmarks found under the label 'LAN'

    Parameters
    ----------
    filepath : `Path`
        Absolute filepath of the file.
    asset : `object`, optional
        An optional asset that may help with loading. This is unused for this
        implementation.
    \**kwargs : `dict`, optional
        Any other keyword arguments.

    Returns
    -------
    landmarks : :map:`LandmarkGroup`
        The landmarks including appropriate labels if available.
    """
    with open(str(filepath), 'r') as f:
        points = np.fromfile(
            f, dtype=np.float32)[3:].reshape([-1, 3]).astype(np.double)
    return PointCloud(points)


def bnd_importer(filepath, asset=None, **kwargs):
    r"""
    Importer for the BND file format for the BU-3DFE dataset. This is a 3D
    landmark type and so it is assumed it only applies to meshes.

    Landmark set label: BND

    Landmark labels:

    +---------------+
    | label         |
    +===============+
    | left_eye      |
    | right_eye     |
    | left_eyebrow  |
    | right_eyebrow |
    | nose          |
    | mouth         |
    | chin          |
    +---------------+

    Parameters
    ----------
    filepath : `Path`
        Absolute filepath of the file.
    asset : `object`, optional
        An optional asset that may help with loading. This is unused for this
        implementation.
    \**kwargs : `dict`, optional
        Any other keyword arguments.

    Returns
    -------
    landmarks : :map:`LabelledPointUndirectedGraph`
        The landmarks including appropriate labels.
    """
    with open(str(filepath), 'r') as f:
        points = f.read()

    # Remove blank lines
    landmark_text = [l for l in points.splitlines() if l.rstrip()]
    landmark_text = [l.split() for l in landmark_text]

    n_points = len(landmark_text)
    points = np.zeros([n_points, 3])
    for i, l in enumerate(landmark_text):
        # Skip the first number as it's an index into the mesh
        points[i, :] = np.array([float(l[1]), float(l[2]), float(l[3])],
                                dtype=np.float)
    labels_to_masks = OrderedDict([
        ('left_eye', _indices_to_mask(n_points, np.arange(8))),
        ('right_eye', _indices_to_mask(n_points, np.arange(8, 16))),
        ('left_eyebrow', _indices_to_mask(n_points, np.arange(16, 26))),
        ('right_eyebrow', _indices_to_mask(n_points, np.arange(26, 36))),
        ('nose', _indices_to_mask(n_points, np.arange(36, 48))),
        ('mouth', _indices_to_mask(n_points, np.arange(48, 68))),
        ('chin', _indices_to_mask(n_points, np.arange(68, 83)))
    ])
    empty_adj_matrix = csr_matrix((n_points, n_points))
    return LabelledPointUndirectedGraph(points, empty_adj_matrix,
                                        labels_to_masks)


def _indices_to_mask(n_points, indices):
    """
    Helper function to turn an array of indices in to a boolean mask.

    Parameters
    ----------
    n_points : `int`
        The total number of points for the mask
    indices : `int` `ndarray`
        An array of integers representing the `True` indices.

    Returns
    -------
    boolean_mask : `bool` `ndarray`
        The mask for the set of landmarks where each index from indices is set
        to `True` and the rest are `False`
    """
    mask = np.zeros(n_points, dtype=np.bool)
    mask[indices] = True
    return mask



from __future__ import division
import numpy as np
import warnings

from menpo.base import Vectorizable, MenpoMissingDependencyError
from menpo.transform import Transform, Translation, Rotation


def align_2d_3d(points_3d, points_image, image_shape, focal_length=None,
                distortion_coeffs=None):
    try:
        import cv2
    except ImportError:
        raise MenpoMissingDependencyError('opencv3')
    height, width = image_shape
    # Create camera matrix
    focal_length = (max(height, width)
                    if focal_length is None else focal_length)
    c_x = width / 2.0
    c_y = height / 2.0
    camera_matrix = np.array([[focal_length, 0, c_x],
                              [0, focal_length, c_y],
                              [0, 0, 1.0]])

    # If distortion coefficients are None, set them to zero
    if distortion_coeffs is None:
        distortion_coeffs = np.zeros(4)
    # Estimate the camera pose given the 3D sparse pointcloud on the
    # mesh, its 2D projection on the image, the camera matrix and the
    # distortion coefficients
    lm2d = points_image.points[:, ::-1].copy()
    converged, r_vec, t_vec = cv2.solvePnP(points_3d.points,
                                           lm2d,
                                           camera_matrix, distortion_coeffs)

    if not converged:
        warnings.warn('cv2.SolvePnP did not converge to a solution')

    # Create rotation and translation transform objects from the vectors
    # acquired at the previous step
    rotation_matrix = cv2.Rodrigues(r_vec)[0]
    r = Rotation(rotation_matrix)
    t = Translation(t_vec.ravel())

    return r, t, focal_length


class OrthographicProjection(Transform, Vectorizable):

    def __init__(self, focal_length, image_shape):
        self.focal_length = focal_length
        self.height = image_shape[0]
        self.width = image_shape[1]

    @property
    def n_dims(self):
        return 3

    @property
    def n_parameters(self):
        return 1

    def as_vector(self):
        return np.array([self.focal_length])

    def _from_vector_inplace(self, vector):
        self.focal_length = vector[0]

    def _apply(self, x, **kwargs):
        f = self.focal_length
        c_x = self.width / 2
        c_y = self.height / 2

        output = np.empty_like(x)
        output[:, 0] = (f * x[:, 1]) + c_y
        output[:, 1] = (f * x[:, 0]) + c_x
        output[:, 2] = x[:, 2]

        return output


class PerspectiveProjection(OrthographicProjection):
    def _apply(self, x, **kwargs):
        f = self.focal_length
        c_x = self.width / 2
        c_y = self.height / 2

        output = np.empty_like(x)
        output[:, 0] = (f * x[:, 1]) / x[:, 2] + c_y
        output[:, 1] = (f * x[:, 0]) / x[:, 2] + c_x
        output[:, 2] = x[:, 2]

        return output

    @classmethod
    def init_from_2d_projected_shape(cls, points_3d, points_image,
                                     image_shape, focal_length=None,
                                     distortion_coeffs=None):
        r, t, focal_length = align_2d_3d(
            points_3d, points_image, image_shape, focal_length=focal_length,
            distortion_coeffs=distortion_coeffs)
        return OrthographicCamera(r, t, OrthographicProjection(focal_length,
                                                               image_shape))

class OrthographicCamera(Vectorizable):
    def __init__(self, rotation, translation, projection):
        self.rotation_transform = rotation
        self.translation_transform = translation
        self.projection_transform = projection

    @property
    def focal_length(self):
        return self.projection_transform.focal_length

    @classmethod
    def init_from_image_shape_and_vector(cls, image_shape, vector):
        r = Rotation.init_identity(n_dims=3)
        t = Translation.init_identity(n_dims=3)
        p = OrthographicProjection(focal_length=1, image_shape=image_shape)
        return cls(r, t, p).from_vector(vector)

    @classmethod
    def init_from_2d_projected_shape(cls, points_3d, points_image,
                                     image_shape, focal_length=None,
                                     distortion_coeffs=None):
        raise NotImplementedError("Orthographic camera pose estimation not "
                                  "implemented.")

    def apply(self, instance, **kwargs):
        return self.camera_transform.apply(instance)

    @property
    def n_parameters(self):
        return (self.projection_transform.n_parameters +
                self.rotation_transform.n_parameters +
                self.translation_transform.n_parameters)

    def as_vector(self):
        # focal_length, q_w, q_x, q_y, q_z, t_x, t_y, t_z
        params = np.zeros(self.n_parameters)

        # focal length
        params[:1] = self.projection_transform.as_vector()

        # 4 parameters: q_w, q_x, q_y, q_z
        params[1:5] = self.rotation_transform.as_vector()

        # 3 parameters: t_x, t_y, t_z
        params[5:] = self.translation_transform.as_vector()
        return params

    def _from_vector_inplace(self, vector):
        self.projection_transform._from_vector_inplace(vector[:1])
        self.rotation_transform._from_vector_inplace(vector[1:5])
        self.translation_transform._from_vector_inplace(vector[5:])

    @property
    def view_transform(self):
        return self.rotation_transform.compose_before(self.translation_transform)

    @property
    def camera_transform(self):
        return self.view_transform.compose_before(self.projection_transform)


class PerspectiveCamera(OrthographicCamera):
    @classmethod
    def init_from_image_shape_and_vector(cls, image_shape, vector):
        r = Rotation.init_identity(n_dims=3)
        t = Translation.init_identity(n_dims=3)
        p = PerspectiveProjection(focal_length=1, image_shape=image_shape)
        return cls(r, t, p).from_vector(vector)

    @classmethod
    def init_from_2d_projected_shape(cls, points_3d, points_image,
                                     image_shape, focal_length=None,
                                     distortion_coeffs=None):
        r, t, focal_length = align_2d_3d(
            points_3d, points_image, image_shape, focal_length=focal_length,
            distortion_coeffs=distortion_coeffs)
        return PerspectiveCamera(r, t, PerspectiveProjection(focal_length,
                                                             image_shape))



from menpo.io.input.landmark import ljson_importer
from .mesh import (wrl_importer, mjson_importer, obj_importer,
                   stl_importer, ply_importer)
from .landmark_mesh import (bnd_importer, lan_importer, lm3_importer,
                            pts_mesh_importer)
from .lsfm import lsfm_model_importer

# TODO: Add PLY (ASCII and binary) and OFF importers
mesh_types = {'.obj': obj_importer,
              '.stl': stl_importer,
              '.ply': ply_importer,
              '.wrl': wrl_importer,
              '.mjson': mjson_importer}

mesh_landmark_types = {'.pts3': pts_mesh_importer,
                       '.lm3': lm3_importer,
                       '.lan': lan_importer,
                       '.bnd': bnd_importer,
                       '.ljson': ljson_importer}

lsfm_types = {
    '.mat': lsfm_model_importer
}



from .base import (import_builtin_asset, mesh_paths,
                   import_landmark_file, import_landmark_files,
                   menpo3d_data_path_to as data_path_to,
                   menpo3d_data_dir_path as data_dir_path,
                   menpo3d_ls_builtin_assets as ls_builtin_assets,
                   landmark_file_paths, import_mesh, import_meshes,
                   register_landmark_importer, register_mesh_importer,
                   import_lsfm_model)



import os
import platform
import site

import versioneer
from setuptools import setup, find_packages, Extension

SYS_PLATFORM = platform.system().lower()
IS_LINUX = 'linux' in SYS_PLATFORM
IS_OSX = 'darwin' == SYS_PLATFORM
IS_WIN = 'windows' == SYS_PLATFORM

# Get Numpy include path without importing it
NUMPY_INC_PATHS = [os.path.join(r, 'numpy', 'core', 'include')
                   for r in site.getsitepackages() if
                   os.path.isdir(os.path.join(r, 'numpy', 'core', 'include'))]
if len(NUMPY_INC_PATHS) == 0:
    try:
        import numpy as np
    except ImportError:
        raise ValueError("Could not find numpy include dir and numpy not installed before build - "
                         "cannot proceed with compilation of cython modules.")
    else:
        # just ask numpy for it's include dir
        NUMPY_INC_PATHS = [np.get_include()]

elif len(NUMPY_INC_PATHS) > 1:
    print("Found {} numpy include dirs: "
          "{}".format(len(NUMPY_INC_PATHS), ', '.join(NUMPY_INC_PATHS)))
    print("Taking first (highest precedence on path): {}".format(
        NUMPY_INC_PATHS[0]))
NUMPY_INC_PATH = NUMPY_INC_PATHS[0]


# ---- C/C++ EXTENSIONS ---- #
# Stolen (and modified) from the Cython documentation:
#     http://cython.readthedocs.io/en/latest/src/reference/compilation.html
def no_cythonize(extensions, **_ignore):
    import os.path as op
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in ('.pyx', '.py'):
                if extension.language == 'c++':
                    ext = '.cpp'
                else:
                    ext = '.c'
                sfile = path + ext
                if not op.exists(sfile):
                    raise ValueError('Cannot find pre-compiled source file '
                                     '({}) - please install Cython'.format(sfile))
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions


def build_extension_from_pyx(pyx_path, extra_sources_paths=None):
    if extra_sources_paths is None:
        extra_sources_paths = []
    extra_sources_paths.insert(0, pyx_path)
    ext = Extension(name=pyx_path[:-4].replace('/', '.'),
                    sources=extra_sources_paths,
                    include_dirs=[NUMPY_INC_PATH],
                    language='c++')
    if IS_LINUX or IS_OSX:
        ext.extra_compile_args.append('-Wno-unused-function')
    if IS_OSX:
        ext.extra_link_args.append('-headerpad_max_install_names')
    return ext


try:
    from Cython.Build import cythonize
except ImportError:
    import warnings

    cythonize = no_cythonize
    warnings.warn('Unable to import Cython - attempting to build using the '
                  'pre-compiled C++ files.')

cython_modules = [
    build_extension_from_pyx('menpo3d/rasterize/tripixel.pyx'),
]
cython_exts = cythonize(cython_modules, quiet=True)

install_requires = ['menpo>=0.9.0,<0.11.0',
                    'mayavi>=4.7.0',
                    'moderngl>=5.5.*,<6.0']

setup(name='menpo3d',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='Menpo library providing tools for 3D Computer Vision research',
      author='James Booth',
      author_email='james.booth08@imperial.ac.uk',
      packages=find_packages(),
      package_data={'menpo3d': ['data/*']},
      install_requires=install_requires,
      tests_require=['pytest>=5.0', 'mock>=3.0'],
      ext_modules=cython_exts
      )
